import toolviper.utils.logger as logger, numcodecs, os, time, warnings
from itertools import cycle
from pathlib import Path
from typing import Any, Dict, Union

import xarray as xr
import zarr


def write_part_keys(
    partitions: Dict[Any, xr.Dataset], outpath: str, compressor: numcodecs.abc.Codec
) -> None:
    """
    Writes an xds with the partition keys.

    Parameters
    ----------
    partitions : Dict[Any, xr.Dataset]
        partitions from a cds
    outpath : str
        path to write a cds
    compressor : numcodecs.abc.Codec
        compressor used for the partition keys variable

    Returns
    -------

    """

    spw_ids, pol_setup_ids, intents = map(list, zip(*partitions.keys()))
    part_keys = xr.Dataset(
        data_vars={
            "spw_ids": spw_ids,
            "pol_setup_ids": pol_setup_ids,
            "intents": intents,
        }
    )

    encoding = dict(zip(list(part_keys.data_vars), cycle([{"compressor": compressor}])))

    out_path = Path(outpath, "partition_keys")
    xr.Dataset.to_zarr(
        part_keys,
        store=out_path,
        mode="w",
        encoding=encoding,
        consolidated=False,
    )
    zarr.consolidate_metadata(out_path)


def write_metainfo(
    outpath: str,
    metainfo: Dict[str, xr.Dataset],
    chunks_on_disk: Union[Dict, None] = None,
    compressor: Union[numcodecs.abc.Codec, None] = None,
    consolidated: bool = True,
) -> None:
    """
    Write all metainfo subtables from a cds to zarr storage

    Parameters
    ----------
    outpath : str

    metainfo : Dict[str, xr.Dataset]:

    chunks_on_disk : Union[Dict, None] (Default value = None)

    compressor : Union[numcodecs.abc.Codec, None) (Default value = None)

    consolidated : bool (Default value = True)


    Returns
    -------

    """
    metadir = Path(outpath, "metainfo")
    os.mkdir(metadir)
    for key, xds in metainfo.items():
        xds_outpath = Path(metadir, key)
        logger.debug(f"Saving metainfo xds {key} into {xds_outpath}")
        write_xds_to_zarr(
            xds, key, xds_outpath, chunks_on_disk, compressor, consolidated=True
        )


def write_partitions(
    outpath: str,
    partitions: Dict[str, xr.Dataset],
    chunks_on_disk: Union[Dict, None] = None,
    compressor: Union[numcodecs.abc.Codec, None] = None,
    consolidated: bool = True,
) -> None:
    """
    Write all data partitions metainfo from a cds to zarr storage

    Parameters
    ----------
    outpath : str :

    partitions : Dict[str, xr.Dataset]

    chunks_on_disk : Union[Dict, None] (Default value = None)

    compressor : Union[numcodecs.abc.Codec, None] (Default value = True)

    consolidated: bool (Default value = True)


    Returns
    -------

    """

    partdir = Path(outpath, "partitions")
    os.mkdir(partdir)
    for cnt, (key, xds) in enumerate(partitions.items()):
        xds_name = f"xds_{cnt}"
        xds_outpath = Path(partdir, str(xds_name))
        logger.debug(f"Saving partition xds {key} into {xds_outpath}")
        write_xds_to_zarr(
            xds, xds_name, xds_outpath, chunks_on_disk, compressor, consolidated=True
        )


def write_xds_to_zarr(
    xds: xr.Dataset,
    name: str,
    outpath: str,
    chunks_on_disk: Union[Dict, None] = None,
    compressor: Union[numcodecs.abc.Codec, None] = None,
    consolidated: bool = True,
    graph_name: str = "write_xds_to_zarr",
) -> None:
    """
    Write one xr dataset from a cds (either metainfo or a partition).

    Parameters
    ----------
    xds : xr.Dataset
        cds (sub)dataset
    name : str
        dataset name (example subtable name, or xds{i})
    outpath: str :

    chunks_on_disk : Union[Dict, None] (Default value = None)

    compressor : Union[numcodecs.abc.Codec, None] (Default value = None)

    consolidated : bool (Default value = True)

    graph_name : str
        the time taken to execute the graph and save the
        dataset is measured and saved as an attribute in the zarr file.
        The graph_name is the label for this timing information.

    Returns
    -------

    """

    xds_for_disk = xds
    if chunks_on_disk is not None:
        xds_for_disk = xds_for_disk.chunk(chunks=chunks_on_disk)

    xds_for_disk = prepare_attrs_for_zarr(name, xds_for_disk)

    if name.startswith("xds"):
        # Do not write replicated/interpolated pointing (keep only original
        # pointing subtable
        pointing_vars = [
            var for var in xds_for_disk.data_vars if var.startswith("pointing_")
        ]
        xds_for_disk = xds_for_disk.drop_vars(pointing_vars)
        # https://github.com/pydata/xarray/issues/2300
        # Could: compute / xds_for_disk[var] = xds_for_disk[var].chunk(chunks_on_disk)

    for var in xds_for_disk.data_vars:
        # Need a dequantify/quantify-kind of pair of functions for units?
        if (
            xds_for_disk.data_vars[var].dtype == "datetime64[ns]"
            and "units" in xds_for_disk.data_vars[var].attrs
        ):
            xds_for_disk.data_vars[var].attrs.pop("units")

    for coord in xds_for_disk.coords:
        if (
            xds_for_disk.coords[coord].dtype == "datetime64[ns]"
            and "units" in xds_for_disk.coords[coord].attrs
        ):
            xds_for_disk.coords[coord].attrs.pop("units")

    # Create compression encoding for each datavariable
    encoding = dict(
        zip(list(xds_for_disk.data_vars), cycle([{"compressor": compressor}]))
    )

    start = time.time()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "invalid value encountered in cast")
        xr.Dataset.to_zarr(
            xds_for_disk,
            store=outpath,
            mode="w",
            encoding=encoding,
            consolidated=consolidated,
        )
    time_to_calc_and_store = time.time() - start
    logger.debug(
        f"time to store and execute graph ({graph_name}) for {name}: {time_to_calc_and_store}"
    )

    # Add timing information?
    # dataset_group = zarr.open_group(xds_outpath, mode='a')
    # dataset_group.attrs[graph_name+'_time'] = time_to_calc_and_store

    if consolidated:
        zarr.consolidate_metadata(outpath)


def prepare_attrs_for_zarr(name: str, xds: xr.Dataset) -> xr.Dataset:
    """
    Deal with types that cannot be serialized as they are in the
    cds/xds (ndarray etc.)

    Parameters
    ----------
    name : str

    xds : xr.Dataset


    Returns
    -------

    """
    ctds_attrs = xds.attrs["other"]["msv2"]["ctds_attrs"]
    col_descrs = ctds_attrs["column_descriptions"]

    # Present in some old vla/alma MSs (model_0, model_1, model_2...
    # vla/ngc5921.applycal.ms, alma/uid___A002_X71a45c_X1d24.ms.split): Just drop.
    for attr in list(ctds_attrs):
        if attr.startswith("model_"):
            ctds_attrs.pop(attr)

    # Seen for example in vla/ngc5921_statwt_ref_test_algorithm_sep_corr_no_fitspw.ms. Just drop.
    data_cols = ["DATA", "CORRECTED_DATA", "MODEL_DATA"]
    for col in data_cols:
        if col in col_descrs and "CHANNEL_SELECTION" in col_descrs[col]["keywords"]:
            col_descrs[col]["keywords"].pop("CHANNEL_SELECTION")

    if "xds" in name:
        col_descrs["UVW"]["shape"] = ",".join(map(str, col_descrs["UVW"]["shape"]))

    if "spectral_window" == name:
        for col in ["CHAN_FREQ", "REF_FREQUENCY"]:
            measinfo = col_descrs[col]["keywords"]["MEASINFO"]
            if "TabRefCodes" in measinfo:
                measinfo["TabRefCodes"] = ",".join(
                    map(
                        str,
                        measinfo["TabRefCodes"],
                    )
                )

    if "antenna" == name:
        for col in ["OFFSET", "POSITION"]:
            if col in col_descrs:
                col_descrs[col]["shape"] = ",".join(map(str, col_descrs[col]["shape"]))
        # ARRAY_POSITION present in keywords of some ALMA-SD datasets
        # example: almasd/expected.ms
        for kw in ["ARRAY_POSITION"]:
            if kw in col_descrs["POSITION"]["keywords"]:
                col_descrs["POSITION"]["keywords"][kw] = ",".join(
                    map(str, col_descrs["POSITION"]["keywords"][kw])
                )

    if "ephemerides" == name:
        if "radii" in ctds_attrs:
            ctds_attrs["radii"]["value"] = ",".join(
                map(str, ctds_attrs["radii"]["value"])
            )

    if "feed" == name:
        col_descrs["POSITION"]["shape"] = ",".join(
            map(str, col_descrs["POSITION"]["shape"])
        )

    if "field" == name:
        for col in ["DELAY_DIR", "PHASE_DIR", "REFERENCE_DIR"]:
            # These shapes are not present in many MSs
            if col in col_descrs and "shape" in col_descrs[col]:
                col_descrs[col]["shape"] = ",".join(map(str, col_descrs[col]["shape"]))
            if "TabRefCodes" in col_descrs[col]["keywords"]["MEASINFO"]:
                col_descrs[col]["keywords"]["MEASINFO"]["TabRefCodes"] = ",".join(
                    map(
                        str,
                        col_descrs[col]["keywords"]["MEASINFO"]["TabRefCodes"],
                    )
                )

    if "observation" == name:
        col_descrs["TIME_RANGE"]["shape"] = ",".join(
            map(str, col_descrs["TIME_RANGE"]["shape"])
        )

    if "source" == name:
        # Note several of these cols are optional and/or only
        # populated with arrays sometimes!
        for col in [
            "DIRECTION",
            "PROPER_MOTION",
            "POSITION",
            "TRANSITION",
            "REST_FREQUENCY",
            "SYSVEL",
        ]:
            if col in col_descrs and "shape" in col_descrs[col]:
                col_descrs[col]["shape"] = ",".join(map(str, col_descrs[col]["shape"]))

    if "weather" == name:
        # Non-std col
        for col in ["NS_WX_STATION_POSITION"]:
            if col in col_descrs and "shape" in col_descrs[col]:
                col_descrs[col]["shape"] = ",".join(map(str, col_descrs[col]["shape"]))

    return xds



import toolviper.utils.logger as logger
from typing import Dict, List, Tuple, Union

import pandas as pd
import numpy as np
import xarray as xr

from casacore import tables

from .load import load_col_chunk
from .read_main_table import get_partition_ids, redim_id_data_vars, rename_vars
from .read import add_units_measures, convert_casacore_time, extract_table_attributes
from .write import revert_time
from .table_query import open_query, open_table_ro
from xradio.measurement_set._utils._ms._tables.read_main_table import (
    get_baselines,
    get_baseline_indices,
)
from xradio._utils.list_and_array import unique_1d


def load_expanded_main_table_chunk(
    infile: str,
    ddi: int,
    chunk: Dict[str, slice],
    ignore_msv2_cols: Union[list, None] = None,
) -> xr.Dataset:
    """
    Load a chunk of data from main table into memory, with expanded
    dims: (time, baseline, freq, pols)

    Parameters
    ----------
    infile : str
        Input MS path
    ddi : int
        DDI to load chunk from
    chunk : Dict[str, slice]
        specification of chunk to load
    ignore_msv2_cols : Union[list, None] (Default value = None)
        cols that should not be loaded (deprecated MSv2 or similar)

    Returns
    -------
    xr.Dataset
        Xarray datasets with chunk of visibility data, one per DDI (spw_id, pol_setup_id) pair
    """

    taql_where = f"where DATA_DESC_ID = {ddi}"
    taql_ddi = f"select * from $mtable {taql_where}"

    with open_table_ro(infile) as mtable:
        with open_query(mtable, taql_ddi) as tb_tool:
            if tb_tool.nrows() == 0:
                return xr.Dataset()

            xds, part_ids, attrs = load_expanded_ddi_chunk(
                infile, tb_tool, taql_where, chunk, ignore_msv2_cols
            )

    return xds, part_ids, attrs


def load_expanded_ddi_chunk(
    infile: str,
    tb_tool: tables.table,
    taql_pre: str,
    chunk: Dict[str, slice],
    ignore_msv2_cols: Union[list, None] = None,
) -> xr.Dataset:
    """
    Helper function to effectively load the chunk and produce an
    xr.Dataset from a DII once the table and initial query(ies) have
    been opened successfully.

    Parameters
    ----------
    infile : str
        Input MS path
    tb_tool : tables.table
        table query contrained to one DDI and chunk time range
    taql_pre : str
        TaQL query used for tb_tool, with some pre-selection of rows and columns
    chunk : Dict[str, slice]
        specification of data chunk to load
    ignore_msv2_cols : Union[list, None] (Default value = None)
        propagated from calling funtions

    Returns
    -------
    xr.Dataset
        An Xarray dataset with data variables as plain numpy
        arrays loaded directly from the MS columns
    """

    # read the specified chunk of data, figure out indices and lens
    utimes, times = get_chunk_times(taql_pre, chunk)
    baselines, blines = get_chunk_baselines(tb_tool, chunk)
    tidxs, bidxs, didxs, taql_where_chunk = get_chunk_data_indices(
        taql_pre, chunk, utimes, times, baselines, blines
    )

    ctlen = min(len(utimes), times[1] - times[0] + 1)
    cblen = min(len(baselines), blines[1] - blines[0] + 1)
    mvars = load_ddi_cols_chunk(
        ctlen, cblen, tidxs, bidxs, didxs, tb_tool, chunk, ignore_msv2_cols
    )

    mcoords = {
        "time": xr.DataArray(convert_casacore_time(utimes[:ctlen]), dims=["time"]),
        "baseline": xr.DataArray(np.arange(cblen), dims=["baseline"]),
    }

    # add xds global attributes
    cc_attrs = extract_table_attributes(infile)
    attrs = {"other": {"msv2": {"ctds_attrs": cc_attrs, "bad_cols": ignore_msv2_cols}}}
    # add per data var attributes
    mvars = add_units_measures(mvars, cc_attrs)
    mcoords = add_units_measures(mcoords, cc_attrs)

    mvars = rename_vars(mvars)
    mvars = redim_id_data_vars(mvars)
    xds = xr.Dataset(mvars, coords=mcoords)

    part_ids = get_partition_ids(tb_tool, taql_where_chunk)

    # needs an ~equivalent to add_partition_attrs?
    return xds, part_ids, attrs


def load_ddi_cols_chunk(
    ctlen: int,
    cblen: int,
    tidxs: np.ndarray,
    bidxs: np.ndarray,
    didxs: np.ndarray,
    tb_tool: tables.table,
    chunk: Dict[str, slice],
    ignore_msv2_cols: Union[list, None] = None,
) -> Dict[str, np.ndarray]:
    """
    For a given chunk and DDI, load all the MSv2 columns

    Parameters
    ----------
    ctlen : int
        length of the time axis/dim of the chunk
    cblen : int
        length of the baseline axis of the chunk
    tidxs : np.ndarray
        time axis indices
    bidxs : np.ndarray
        baseline axis indices
    didxs : np.ndarray
        (effective) data indices, excluding missing baselines
    tb_tool : tables.table
        a table/TaQL query open and being used to load columns
    chunk : Dict[str, slice]
        data chunk specification
    ignore_msv2_cols : Union[list, None] (Default value = None)
        propagated from calling funtions

    Returns
    -------
    Dict[str, np.ndarray]
        columns loaded into memory as np arrays

    """
    cols = tb_tool.colnames()

    cshapes = [
        np.array(tb_tool.getcell(col, 0)).shape
        for col in cols
        if tb_tool.iscelldefined(col, 0)
    ]
    # Assumes shapes are consistent across columns - MSv2
    chan_cnt, pol_cnt = [(csh[0], csh[1]) for csh in cshapes if len(csh) == 2][0]

    dims = ["time", "baseline", "freq", "pol"]
    mvars = {}
    # loop over each column and load data
    for col in cols:
        if (col in ignore_msv2_cols + ["TIME"]) or not tb_tool.iscelldefined(col, 0):
            continue

        cdata = tb_tool.getcol(col, 0, 1)[0]
        cell_shape = cdata.shape
        if len(cell_shape) == 0:
            col_dims = dims[:2]
            mvars[col] = xr.DataArray(
                load_col_chunk(
                    tb_tool, col, (ctlen, cblen), tidxs, bidxs, didxs, None, None
                ),
                dims=col_dims,
            )

        elif col == "UVW":
            col_dims = dims[:2] + ["uvw_coords"]
            mvars[col] = xr.DataArray(
                load_col_chunk(
                    tb_tool, col, (ctlen, cblen, 3), tidxs, bidxs, didxs, None, None
                ),
                dims=col_dims,
            )

        elif len(cell_shape) == 1:
            pols, col_dims = get_col_1d_pols(cell_shape, dims, chan_cnt, pol_cnt, chunk)
            cshape = (ctlen, cblen) + (pols[1] - pols[0] + 1,)
            mvars[col] = xr.DataArray(
                load_col_chunk(tb_tool, col, cshape, tidxs, bidxs, didxs, pols, None),
                dims=col_dims,
            )

        elif len(cell_shape) == 2:
            chans, pols = get_col_2d_chans_pols(cell_shape, chan_cnt, pol_cnt, chunk)
            cshape = (ctlen, cblen) + (chans[1] - chans[0] + 1, pols[1] - pols[0] + 1)
            col_dims = dims
            mvars[col] = xr.DataArray(
                load_col_chunk(tb_tool, col, cshape, tidxs, bidxs, didxs, chans, pols),
                dims=col_dims,
            )

    return mvars


def get_chunk_times(
    taql_pre: str, chunk: Dict[str, slice]
) -> Tuple[np.ndarray, Tuple[int, int]]:
    """
    Produces time col/axis related values for a chunk: unique times,
    start/stop times.

    Parameters
    ----------
    taql_pre : str
        TaQL query used for tb_tool, with some pre-selection
        of rows and columns.
    chunk : Dict[str, slice]
        specification of data chunk to load

    Returns
    -------
    Tuple[np.ndarray, Tuple[int, int]]
        array of unique times + (firsr, last) time in the chunk
    """

    taql_utimes = f"select DISTINCT TIME from $mtable {taql_pre}"
    with open_query(None, taql_utimes) as query_utimes:
        utimes = unique_1d(query_utimes.getcol("TIME", 0, -1))
        # add a tol around the time ranges returned by taql
        if len(utimes) < 2:
            tol = 1e-5
        else:
            tol = np.diff(utimes).min() / 4

    if "time" in chunk:
        time_slice = chunk["time"]
        if (
            type(time_slice.start) == pd.Timestamp
            and type(time_slice.stop) == pd.Timestamp
        ):
            times = (
                revert_time(time_slice.start) - tol,
                revert_time(time_slice.stop) + tol,
            )
        elif (
            int(time_slice.start) == time_slice.start
            and int(time_slice.stop) == time_slice.stop
        ):
            # could be operator.index(time_slice.start):
            nutimes = len(utimes)
            times = (
                min(nutimes, int(time_slice.start)),
                min(nutimes, int(time_slice.stop)) - 1,
            )
        else:
            raise ValueError(
                f"Invalid time type. Not a timestamp and Cannot use as"
                f" index: {time_slice.start} (type: {type(time_slice.start)})"
            )
    else:
        times = (utimes[0], utimes[-1])

    return utimes, times


def get_chunk_baselines(
    tb_tool: tables.table, chunk: Dict[str, slice]
) -> Tuple[np.ndarray, Tuple[int, int]]:
    """
    Produces the basline col/axis related values for a chunk: an array of
    baselines and the start/stop baseline indices.

    Parameters
    ----------
    tb_tool : tables.table
        table/query opened with prev selections (time)
    chunk : Dict[str, slice]
        specification of data chunk to load

    Returns
    -------
    Tuple[np.ndarray, Tuple[int, int]]
        array of baselines + (first, last) baseline in the chunk
    """
    baselines = get_baselines(tb_tool)

    if "baseline" in chunk:
        baseline_chunk = chunk["baseline"]
        baseline_boundaries = (int(baseline_chunk.start), int(baseline_chunk.stop))
    else:
        baseline_boundaries = (baselines[0][0], baselines[-1][0] - 1)

    return baselines, baseline_boundaries


def get_chunk_data_indices(
    taql_pre: str,
    chunk: Dict[str, slice],
    utimes: np.ndarray,
    times: Tuple[int, int],
    baselines: np.ndarray,
    blines: Tuple[int, int],
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, str]:
    """
    Produces indices to pass to the casacore getcol(slice) functions to load
    the chunk of data. tidxs (time), bidxs (baseline), didxs (effective data
    indices, considering present/absent baselines).

    Time selection is added on top of that.

    Parameters
    ----------
    taql_pre : str
        TaQL query constraints to prepend/inject
    chunk : Dict[str, slice]
        specification of data chunk
    utimes : np.ndarray
        array of times in the chunk
    times : Tuple[int, int]
        start, stop time indices
    baselines : np.ndarray
        array of baselines inthe chunk
    blines : Tuple[int, int]
        start, stop baseline indices

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray]
        indices along the time, baseline and data (time/baseline)
        axes + the full where... string defined for this chunk

    """

    taql_time = f"TIME BETWEEN {utimes[times[0]]} AND {utimes[times[1]]}"
    taql_ant = f"ANTENNA1 BETWEEN {blines[0]} and {blines[1]}"
    taql_where_chunk = f"{taql_pre} AND {taql_time} AND {taql_ant}"
    taql_chunk = f"select * from $mtable {taql_where_chunk}"
    with open_query(None, taql_chunk) as query_times_ants:
        logger.debug(
            f"Opened chunk query, with {query_times_ants.nrows()} rows. Query: {taql_chunk}"
        )
        tidxs = (
            np.searchsorted(utimes, query_times_ants.getcol("TIME", 0, -1)) - times[0]
        )
        ts_ant1, ts_ant2 = (
            query_times_ants.getcol("ANTENNA1", 0, -1),
            query_times_ants.getcol("ANTENNA2", 0, -1),
        )

        ts_bases = np.column_stack((ts_ant1, ts_ant2))

        bidxs = get_baseline_indices(baselines, ts_bases) - blines[0]

    # some antenna 2's will be out of bounds for this chunk, store rows that are in bounds
    didxs = np.where(
        (bidxs >= 0)
        & (bidxs < min(blines[1] - blines[0] + 1, len(baselines) - blines[0]))
    )[0]

    return tidxs, bidxs, didxs, taql_where_chunk


def get_col_1d_pols(
    cell_shape: Tuple[int],
    dims: List[str],
    chan_cnt: int,
    pol_cnt: int,
    chunk: Dict[str, slice],
) -> Tuple[Tuple[int, int], List[str]]:
    """
    For a column with 1d array values, calculate the start/stop
    indices for the last dimension (either pol or freq).
    It also produces the adequate dimension names.

    Parameters
    ----------
    cell_shape : Tuple[int]
        shape of the column
    dims : List[str]
        full list of dataset dimensions
    chan_cnt : int
        number of channels
    pol_cnt : int
        number of pols
    chunk : Dict[str, slice]
        data chunk specification

    Returns
    -------
    Tuple[Tuple[int, int], List[str]]
        first and last pol/freq index of the chunk, and its
        dimension names

    """
    if cell_shape == chan_cnt:
        # chan/freq
        col_dims = dims[:2] + ["freq"]
        if "freq" in chunk:
            pols = (
                min(chan_cnt, chunk["freq"].start),
                min(chan_cnt, chunk["freq"].stop) - 1,
            )
        else:
            pols = (0, cell_shape[0])
    else:
        # pol
        col_dims = dims[:2] + ["pol"]
        if "pol" in chunk:
            pols = (
                min(pol_cnt, chunk["pol"].start),
                min(pol_cnt, chunk["pol"].stop) - 1,
            )
        else:
            pols = (0, cell_shape[0])

    return pols, col_dims


def get_col_2d_chans_pols(
    cell_shape: Tuple[int],
    chan_cnt: int,
    pol_cnt: int,
    chunk: Dict[str, slice],
) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    """
    For a column with 2d array values (FLAG, DATA, WEIGHT_SPECTRUM,
    etc., calculate the the start/stop indices for the last two
    dimensions of the chunk (freq and pol).
    The dimension names can be assumed to be the full list of dims in
    visibilities (time, baseline, freq, pol).

    Parameters
    ----------
    cell_shape : Tuple[int]
        shape of the column
    chan_cnt : int
        number of channels
    pol_cnt : int
        number of pols
    chunk : Dict[str, slice]
        data chunk specification

    Returns
    -------
    Tuple[Tuple[int, int], Tuple[int, int]]
        first and last index for freq (channel) and pol axes of
        the chunk

    """
    if "freq" in chunk:
        chans = (
            min(chan_cnt, chunk["freq"].start),
            min(chan_cnt, chunk["freq"].stop) - 1,
        )
    else:
        chans = (0, cell_shape[0])

    if "pol" in chunk:
        pols = (
            min(pol_cnt, chunk["pol"].start),
            min(pol_cnt, chunk["pol"].stop) - 1,
        )
    else:
        pols = (0, cell_shape[1])

    return chans, pols



import dataclasses
import typing
import inspect
import functools

import xarray
import numpy
from typeguard import check_type, TypeCheckError

from xradio.schema import (
    metamodel,
    bases,
    xarray_dataclass_to_array_schema,
    xarray_dataclass_to_dataset_schema,
    xarray_dataclass_to_dict_schema,
)


@dataclasses.dataclass
class SchemaIssue:
    """
    Representation of an issue found in a schema check

    As schemas can be quite big, ``path`` can be used to precisely locate the
    source of the issue.
    """

    path: typing.List[typing.Tuple[str, str]]
    """Path to offending data item, using pairs of (entity type, entity name).
    Entity types can be ``data_var``, ``coord`` or ``attr``.

    Example: ``[('data_var', 'foo'), ('coord','bar'), ('attr', 'asd')]``
    refers to ``obj.data_vars['foo'].cords['bar'].attrs['asd']``
    """
    message: str
    """
    Explanation of the issue
    """
    found: typing.Optional[typing.Any] = None
    """
    What was found. Can be any type (type, dtype, value)
    """
    expected: typing.Optional[typing.List[typing.Any]] = None
    """
    List of expected values. Can be any type (type, dtype, value)
    """

    def path_str(self):
        strs = []
        for path, ix in self.path:
            if ix is None or ix == "":
                strs.append(path)
            else:
                strs.append(f"{path}['{ix}']")
        return ".".join(strs)

    def __repr__(self):
        err = f"Schema issue with {self.path_str()}: {self.message}"
        if self.expected is not None:
            options = " or ".join(repr(option) for option in self.expected)
            if self.found is not None:
                err += f" (expected: {options} found: {repr(self.found)})"
            else:
                err += f" (expected: {options})"
        return err


class SchemaIssues(Exception):
    """
    List of issues found in a schema check

    Can be thrown as an exception, so we can report on multiple schema issues
    in one go.
    """

    issues: [SchemaIssue]
    """List of issues found"""

    def __init__(self, issues=None):
        if issues is None:
            self.issues = []
        elif isinstance(issues, SchemaIssues):
            self.issues = issues.issues
        else:
            self.issues = list(issues)

    def at_path(self, elem: str, ix: typing.Optional[str] = None) -> "SchemaIssues":
        for issue in self.issues:
            issue.path.insert(0, (elem, ix))
        return self

    def add(self, issue: SchemaIssue):
        self.issues.append(issue)

    def __iadd__(self, other: "SchemaIssues"):
        self.issues += other.issues
        return self

    def __add__(self, other: "SchemaIssues"):
        new_issues = SchemaIssues(self)
        new_issues += other
        return new_issues

    def __len__(self):
        return len(self.issues)

    def __getitem__(self, ix):
        return self.issues[ix]

    def __str__(self):
        if not self.issues:
            return "No schema issues found"
        else:
            issues_string = "\n * ".join(repr(issue) for issue in self.issues)
            return f"\n * {issues_string}"

    def __repr__(self):
        return f"SchemaIssues({str(self)})"

    def expect(
        self, elem: typing.Optional[str] = None, ix: typing.Optional[str] = None
    ):
        """
        Raises this object if issues were found

        :param elem: If given, will be added to path
        :param ix: If given, will be added to path
        :raises: SchemaIssues
        """
        if elem is not None:
            self.at_path(elem, ix)

        # Hide this function in pytest tracebacks
        __tracebackhide__ = True
        if self.issues:
            raise self


def check_array(
    array: xarray.DataArray, schema: typing.Union[type, metamodel.ArraySchema]
) -> SchemaIssues:
    """
    Check whether an xarray DataArray conforms to a schema

    :param array: DataArray to check
    :param schema: Schema to check against (possibly as :py:class:`AsDataset`)
    :returns: List of :py:class:`SchemaIssue`s found
    """

    # Check that this is actually a DataArray
    if not isinstance(array, xarray.DataArray):
        raise TypeError(
            f"check_array: Expected xarray.DataArray, but got {type(array)}!"
        )
    if bases.is_dataarray_schema(schema):
        schema = xarray_dataclass_to_array_schema(schema)
    if not isinstance(schema, metamodel.ArraySchema):
        raise TypeError(f"check_array: Expected ArraySchema, but got {type(schema)}!")

    # Check dimensions
    issues = check_dimensions(array.dims, schema.dimensions)

    # Check type
    issues += check_dtype(array.dtype, schema.dtypes)

    # Check attributes
    issues += check_attributes(array.attrs, schema.attributes)

    # Check coordinates
    issues += check_data_vars(array.coords, schema.coordinates, "coords")

    return issues


def check_dataset(
    dataset: xarray.Dataset, schema: typing.Union[type, metamodel.DatasetSchema]
) -> SchemaIssues:
    """
    Check whether an xarray DataArray conforms to a schema

    :param array: DataArray to check
    :param schema: Schema to check against (possibly as :py:class:`AsDataArray` dataclass)
    :returns: List of :py:class:`SchemaIssue`s found
    """

    # Check that this is actually a Dataset
    if not isinstance(dataset, xarray.Dataset):
        raise TypeError(
            f"check_dataset: Expected xarray.Dataset, but got {type(dataset)}!"
        )
    if bases.is_dataset_schema(schema):
        schema = xarray_dataclass_to_dataset_schema(schema)
    if not isinstance(schema, metamodel.DatasetSchema):
        raise TypeError(
            f"check_dataset: Expected DatasetSchema, but got {type(schema)}!"
        )

    # Check dimensions. Order does not matter on datasets
    issues = check_dimensions(dataset.dims, schema.dimensions, check_order=False)

    # Check attributes
    issues += check_attributes(dataset.attrs, schema.attributes)

    # Check coordinates
    issues += check_data_vars(dataset.coords, schema.coordinates, "coords")

    # Check data variables
    issues += check_data_vars(dataset.data_vars, schema.data_vars, "data_vars")

    return issues


def check_dimensions(
    dims: [str], expected: [[str]], check_order: bool = True
) -> SchemaIssues:
    """
    Check whether a dimension list conforms to a schema

    :param array: Dimension list to check
    :param schema: Expected possibilities for dimension list
    :param check_order: Whether to check order of dimensions
    :returns: List of :py:class:`SchemaIssue`s found
    """

    # Find a dimension list that matches
    dims_set = set(dims)
    best = None
    best_diff = 0
    for exp_dims in expected:
        exp_dims_set = set(exp_dims)

        # No match? Continue, but take note of best match
        if exp_dims_set != dims_set:
            diff = len(dims_set.symmetric_difference(exp_dims_set))
            if best is None or diff < best_diff:
                best = exp_dims_set
                best_diff = diff
            continue

        # Check that the order matches
        if not check_order or tuple(exp_dims) == tuple(dims):
            return SchemaIssues()

        return SchemaIssues(
            [
                SchemaIssue(
                    path=[("dims", None)],
                    message="Dimensions are in the wrong order! Consider transpose()",
                    found=list(dims),
                    expected=[exp_dims],
                )
            ]
        )

    # Dimensionality not supported - try to give a helpful suggestion
    hint_remove = [f"'{hint}'" for hint in dims_set - best]
    hint_add = [f"'{hint}'" for hint in best - dims_set]
    if hint_remove and hint_add:
        message = f"Unexpected coordinates, replace {','.join(hint_remove)} by {','.join(hint_add)}?"
    elif hint_remove:
        message = f"Superfluous coordinate {','.join(hint_remove)}?"
    elif hint_add:
        message = f"Missing dimension {','.join(hint_add)}!"
    else:
        message = f"Unexpected dimensions/coordinates!"
    return SchemaIssues(
        [
            SchemaIssue(
                path=[("dims", None)],
                message=message,
                found=list(dims),
                expected=list(expected),
            )
        ]
    )


def check_dtype(dtype: numpy.dtype, expected: [numpy.dtype]) -> SchemaIssues:
    """
    Check whether a numpy dtype conforms to a schema

    :param dtype: Numeric type to check
    :param schema: Expected possibilities for dtype
    :returns: List of :py:class:`SchemaIssue`s found
    """

    for exp_dtype in expected:
        # If the expected dtype has no size (e.g. "U", a.k.a. a string of
        # arbitrary length), we don't check itemsize, only kind.
        if (
            dtype.kind == exp_dtype.kind
            and exp_dtype.itemsize == 0
            or exp_dtype == dtype
        ):
            return SchemaIssues()

    # Not sure there's anything more helpful that we can do here? Any special
    # cases worth considering?
    return SchemaIssues(
        [
            SchemaIssue(
                path=[("dtype", None)],
                message="Wrong numpy dtype",
                found=dtype,
                expected=list(expected),
            )
        ]
    )


def check_attributes(
    attrs: typing.Dict[str, typing.Any],
    attrs_schema: typing.List[metamodel.AttrSchemaRef],
    attr_kind: str = "attrs",
) -> SchemaIssues:
    """
    Check whether an attribute set conforms to a schema

    :param attrs: Dictionary of attributes
    :param attrs_schema: Expected schemas
    :returns: List of :py:class:`SchemaIssue`s found
    """

    issues = SchemaIssues()
    for attr_schema in attrs_schema:
        # Attribute missing? Note that a value of "None" is equivalent for the
        # purpose of the check
        val = attrs.get(attr_schema.name)
        if val is None:
            if not attr_schema.optional:
                # Get options
                if typing.get_origin(attr_schema.typ) is typing.Union:
                    options = typing.get_args(attr_schema.typ)
                else:
                    options = [attr_schema.typ]

                issues.add(
                    SchemaIssue(
                        path=[(attr_kind, attr_schema.name)],
                        message=f"Required attribute {attr_schema.name} is missing!",
                        expected=options,
                    )
                )
            continue

        # Check attribute value
        issues += _check_value_union(val, attr_schema.typ).at_path(
            attr_kind, attr_schema.name
        )

    # Extra attributes are always okay

    return issues


def check_data_vars(
    data_vars: typing.Dict[str, xarray.DataArray],
    data_vars_schema: typing.List[metamodel.ArraySchemaRef],
    data_var_kind: str,
) -> SchemaIssues:
    """
    Check whether an data variable set conforms to a schema

    As data variables are data arrays, this will recurse into checking the
    array schemas

    :param data_vars: Dictionary(-like) of data_varinates
    :param data_vars_schema: Expected schemas
    :param datavar_kind: Either 'coords' or 'data_vars'
    :returns: List of :py:class:`SchemaIssue`s found
    """

    assert data_var_kind in ["coords", "data_vars"]

    issues = SchemaIssues()
    for data_var_schema in data_vars_schema:

        allow_mutiple_versions = False
        for attr in data_var_schema.attributes:
            if hasattr(attr, "name"):
                if attr.name == "allow_mutiple_versions":
                    allow_mutiple_versions = attr.default

        data_vars_names = []
        if allow_mutiple_versions:
            for data_var_name in data_vars:
                if data_var_schema.name in data_var_name:
                    data_vars_names.append(data_var_name)
        else:
            data_vars_names = [data_var_schema.name]

        if (len(data_vars_names) == 0) and ~data_var_schema.optional:
            data_vars_names = [data_var_schema.name]

        for data_var_name in data_vars_names:
            data_var = data_vars.get(data_var_name)

            if data_var is None:
                if not data_var_schema.optional:
                    if data_var_kind == "coords":
                        message = (
                            f"Required coordinate '{data_var_schema.name}' is missing!"
                        )
                    else:
                        message = (
                            f"Required data variable '{data_var_schema.name}' is missing "
                            f"(have {','.join(data_vars)})!"
                        )
                    issues.add(
                        SchemaIssue(
                            path=[(data_var_kind, data_var_schema.name)],
                            message=message,
                        )
                    )
                continue

            # Check array schema
            issues += check_array(data_var, data_var_schema).at_path(
                data_var_kind, data_var_schema.name
            )

    # Extra data_varinates / data variables are always okay

    return issues


def check_dict(
    dct: dict, schema: typing.Union[type, metamodel.DictSchema]
) -> SchemaIssues:
    # Check that this is actually a dictionary
    if not isinstance(dct, dict):
        raise TypeError(f"check_dict: Expected dictionary, but got {type(dct)}!")
    if bases.is_dict_schema(schema):
        schema = xarray_dataclass_to_dict_schema(schema)
    if not isinstance(schema, metamodel.DictSchema):
        raise TypeError(f"check_dict: Expected DictSchema, but got {type(schema)}!")

    # Check attributes
    return check_attributes(dct, schema.attributes, attr_kind="")


def _check_value(val, ann):
    """
    Check whether value satisfies annotation

    If the annotation is a data array or dataset schema, it will be checked.

    :param val: Value to check
    :param ann: Type annotation of value
    :returns: Schema issues
    """

    # Is supposed to be a data array?
    if bases.is_dataarray_schema(ann):
        # Attempt to convert dictionaries automatically
        if isinstance(val, dict):
            try:
                val = xarray.DataArray.from_dict(val)
            except ValueError as e:
                return SchemaIssues(
                    [
                        SchemaIssue(
                            path=[], message=str(e), expected=[ann], found=type(val)
                        )
                    ]
                )
            except TypeError as e:
                return SchemaIssues(
                    [
                        SchemaIssue(
                            path=[], message=str(e), expected=[ann], found=type(val)
                        )
                    ]
                )

        if not isinstance(val, xarray.DataArray):
            # Fall through to plain type check
            ann = xarray.DataArray
        else:
            return check_array(val, ann)

    # Is supposed to be a dataset?
    if bases.is_dataset_schema(ann):
        # Attempt to convert dictionaries automatically
        if isinstance(val, dict):
            try:
                val = xarray.Dataset.from_dict(val)
            except ValueError as e:
                return SchemaIssues(
                    [
                        SchemaIssue(
                            path=[], message=str(t), expected=[ann], found=type(val)
                        )
                    ]
                )
        if not isinstance(val, xarray.Dataset):
            # Fall through to plain type check
            ann = xarray.Dataset
        else:
            return check_dataset(val, ann)

    # Is supposed to be a dictionary?
    if bases.is_dict_schema(ann):
        if not isinstance(val, dict):
            # Fall through to plain type check
            ann = dict
        else:
            return check_dict(val, ann)

    # Otherwise straight type check using typeguard
    try:
        check_type(val, ann)
    except TypeCheckError as t:
        return SchemaIssues(
            [SchemaIssue(path=[], message=str(t), expected=[ann], found=type(val))]
        )

    return SchemaIssues()


def _check_value_union(val, ann):
    """
    Check whether value satisfies annotations, including union types

    If the annotation is a data array or dataset schema, it will be checked.

    :param val: Value to check
    :param ann: Type annotation of value
    :returns: Schema issues
    """

    if ann is None or ann is inspect.Signature.empty:
        return SchemaIssues()

    # Account for union types (this especially catches "Optional")
    if typing.get_origin(ann) is typing.Union:
        options = typing.get_args(ann)
    else:
        options = [ann]

    # Go through options, try to find one without issues
    args_issues = None
    okay = False
    for option in options:
        arg_issues = _check_value(val, option)
        # We can immediately return if we find no issues with
        # some schema check
        if not arg_issues:
            return SchemaIssues()
        if args_issues is None:
            args_issues = arg_issues

        # Crude merging of expected options (for "unexpected type")
        elif len(args_issues) == 1 and len(arg_issues) == 1:
            args_issues[0].expected += arg_issues[0].expected

    # Return representative issues list
    if not args_issues:
        raise ValueError("Empty union set?")
    return args_issues


def schema_checked(fn, check_parameters: bool = True, check_return: bool = True):
    """
    Function decorator to check parameters and return value for
    schema conformance

    :param fn: Function to decorate
    :param check_parameters: Whether to check parameters. Can also
       pass an iterable with parameters to check
    :param check_return: Whether to check return value
    :returns: Decorated function
    """
    anns = inspect.getfullargspec(fn).annotations
    signature = inspect.signature(fn)

    if isinstance(check_parameters, bool):
        if check_parameters:
            parameters_to_check = set(signature.parameters)
        else:
            parameters_to_check = {}
    else:
        parameters_to_check = set(check_parameters)
        assert parameters_to_check.issubset(signature.parameters)

    @functools.wraps(fn)
    def _check_fn(*args, **kwargs):
        # Hide this function in pytest tracebacks
        __tracebackhide__ = True

        # Bind parameters, collect (potential) issues
        bound = signature.bind(*args, **kwargs)
        issues = SchemaIssues()
        for arg, val in bound.arguments.items():
            if arg not in parameters_to_check:
                continue

            # Get annotation
            issues += _check_value_union(val, anns.get(arg)).at_path(arg)

        # Any issues found? raise
        issues.expect()

        # Execute function
        result = fn(*args, **kwargs)

        # Check return
        if check_return:
            issues = _check_value_union(val, signature.return_annotation)
            issues.at_path("return").expect()

        # Check return value
        return result

    return _check_fn



from dataclasses import dataclass
from typing import Any, Dict, NamedTuple, Tuple

import numpy as np
import xarray as xr

PartitionKey = Tuple[Any, ...]
# for intent-ddi partitioning
PartitionKeyAsNT = NamedTuple(
    "PartitionKey", [("spw_id", np.int32), ("pol_setup_id", np.int32), ("intent", str)]
)
VisSetPartitions = Dict[PartitionKey, xr.Dataset]
VisSetMetainfo = Dict[str, xr.Dataset]


@dataclass(frozen=True)
class CASAVisSet:
    metainfo: VisSetMetainfo
    partitions: VisSetPartitions
    descr: str

    def __getitem__(self, key):
        return getattr(self, key)

    def __repr__(self):
        """
        Concise repr for CASA vis sets
        """
        class_name = type(self).__name__
        return (
            f"({class_name}\n metainfo (keys)={self.metainfo.keys()}"
            f"\n partitions (keys)={self.partitions.keys()}\n descr={self.descr!r})"
        )

    def _repr_html_(self):
        class_name = type(self).__name__
        return (
            f"({class_name}<p> <b>metainfo</b> (keys)={self.metainfo.keys()}</p><p>"
            f"<b>partitions</b> (keys)={self.partitions.keys()}</p><p> <b>descr</b>={self.descr!r})</p>"
        )



#################################
# Helper File
#
# Not exposed in API
#
#################################
import os
import warnings
from typing import Union

import xarray as xr

from casacore import tables
from ._casacore.common import _open_image_ro
from ._casacore.xds_from_casacore import (
    _add_mask,
    _add_sky_or_aperture,
    _casa_image_to_xds_attrs,
    _casa_image_to_xds_coords,
    _get_mask_names,
    _get_persistent_block,
    _get_starts_shapes_slices,
    _get_transpose_list,
    _multibeam_array,
    _read_image_array,
)
from ._casacore.xds_to_casacore import (
    _coord_dict_from_xds,
    _history_from_xds,
    _imageinfo_dict_from_xds,
    _write_casa_data,
)
from .common import _aperture_or_sky, _get_xds_dim_order, _dask_arrayize_dv

warnings.filterwarnings("ignore", category=FutureWarning)


def _load_casa_image_block(infile: str, block_des: dict, do_sky_coords) -> xr.Dataset:
    image_full_path = os.path.expanduser(infile)
    with _open_image_ro(image_full_path) as casa_image:
        coords = casa_image.coordinates()
        cshape = casa_image.shape()
    ret = _casa_image_to_xds_coords(image_full_path, False, do_sky_coords)
    xds = ret["xds"].isel(block_des)
    starts, shapes, slices = _get_starts_shapes_slices(block_des, coords, cshape)
    dimorder = _get_xds_dim_order(ret["sphr_dims"])
    transpose_list, new_axes = _get_transpose_list(coords)
    block = _get_persistent_block(
        image_full_path, shapes, starts, dimorder, transpose_list, new_axes
    )
    xds = _add_sky_or_aperture(xds, block, dimorder, image_full_path, ret["sphr_dims"])
    mymasks = _get_mask_names(image_full_path)
    for m in mymasks:
        full_path = os.sep.join([image_full_path, m])
        block = _get_persistent_block(
            full_path, shapes, starts, dimorder, transpose_list, new_axes
        )
        xds = _add_mask(xds, m, block, dimorder)
    xds.attrs = _casa_image_to_xds_attrs(image_full_path, True)
    mb = _multibeam_array(xds, image_full_path, False)
    if mb is not None:
        selectors = {}
        for k in ("time", "polarization", "frequency"):
            if k in block_des:
                selectors[k] = block_des[k]
        xds["beam"] = mb.isel(selectors)
    return xds


def _read_casa_image(
    infile: str,
    chunks: Union[list, dict],
    verbose: bool,
    do_sky_coords: bool,
    masks: bool = True,
    history: bool = True,
) -> xr.Dataset:
    img_full_path = os.path.expanduser(infile)
    ret = _casa_image_to_xds_coords(img_full_path, verbose, do_sky_coords)
    xds = ret["xds"]
    dimorder = _get_xds_dim_order(ret["sphr_dims"])
    xds = _add_sky_or_aperture(
        xds,
        _read_image_array(img_full_path, chunks, verbose=verbose),
        dimorder,
        img_full_path,
        ret["sphr_dims"],
    )
    if masks:
        mymasks = _get_mask_names(img_full_path)
        for m in mymasks:
            ary = _read_image_array(img_full_path, chunks, mask=m, verbose=verbose)
            xds = _add_mask(xds, m, ary, dimorder)
    xds.attrs = _casa_image_to_xds_attrs(img_full_path, history)
    mb = _multibeam_array(xds, img_full_path, True)
    if mb is not None:
        xds["beam"] = mb
    # xds = _add_coord_attrs(xds, ret["icoords"], ret["dir_axes"])
    xds = _dask_arrayize_dv(xds)
    return xds


def _xds_to_casa_image(xds: xr.Dataset, imagename: str) -> None:
    image_full_path = os.path.expanduser(imagename)
    _write_casa_data(xds, image_full_path)
    # create coordinates
    ap_sky = _aperture_or_sky(xds)
    coord = _coord_dict_from_xds(xds)
    ii = _imageinfo_dict_from_xds(xds)
    units = xds[ap_sky].attrs["units"] if "units" in xds[ap_sky].attrs else None
    miscinfo = (
        xds.attrs["user"]
        if "user" in xds.attrs and len(xds.attrs["user"]) > 0
        else None
    )
    tb = tables.table(
        image_full_path,
        readonly=False,
        lockoptions={"option": "permanentwait"},
        ack=False,
    )
    tb.putkeyword("coords", coord)
    tb.putkeyword("imageinfo", ii)
    if units:
        tb.putkeyword("units", units)
    if miscinfo:
        tb.putkeyword("miscinfo", miscinfo)
    tb.done()
    # history
    _history_from_xds(xds, image_full_path)



import pandas as pd
from xradio._utils.list_and_array import to_list
import xarray as xr
import numbers
import os
from collections.abc import Mapping, Iterable
from typing import Any, Union


class MeasurementSetXds(xr.Dataset):
    __slots__ = ()

    def __init__(self, xds):
        super().__init__(xds.data_vars, xds.coords, xds.attrs)

    def to_store(self, store, **kwargs):
        """
        Write the MeasurementSetXds to a Zarr store.
        Does not write to cloud storage yet.

        Args:
            store (str): The path to the Zarr store.
            **kwargs: Additional keyword arguments to be passed to `xarray.Dataset.to_zarr`. See https://docs.xarray.dev/en/latest/generated/xarray.Dataset.to_zarr.html for more information.

        Returns:
            None
        """

        copy_cor_xds = self.copy()  # No deep copy

        # Remove field_and_source_xds from all correlated_data (VISIBILITY/SPECTRUM) data variables
        # and save them as separate zarr files.
        for data_group_name, data_group in self.attrs["data_groups"].items():
            del copy_cor_xds[data_group["correlated_data"]].attrs[
                "field_and_source_xds"
            ]

            # print("data_group_name", data_group_name)
            xr.Dataset.to_zarr(
                self[data_group["correlated_data"]].attrs["field_and_source_xds"],
                os.path.join(store, "field_and_source_xds_" + data_group_name),
                **kwargs,
            )

        # Remove xds attributes from copy_cor_xds and save xds attributes as separate zarr files.
        for attrs_name in self.attrs:
            if "xds" in attrs_name:
                del copy_cor_xds.attrs[attrs_name]
                xr.Dataset.to_zarr(
                    self.attrs[attrs_name], os.path.join(store, attrs_name), **kwargs
                )

        # Save copy_cor_xds as zarr file.
        xr.Dataset.to_zarr(
            copy_cor_xds, os.path.join(store, "correlated_xds"), **kwargs
        )

    def sel(
        self,
        indexers: Union[Mapping[Any, Any], None] = None,
        method: Union[str, None] = None,
        tolerance: Union[int, float, Iterable[Union[int, float]], None] = None,
        drop: bool = False,
        **indexers_kwargs: Any,
    ):
        """
        Select data along dimension(s) by label. Overrides `xarray.Dataset.sel <https://xarray.pydata.org/en/stable/generated/xarray.Dataset.sel.html>`__ so that a data group can be selected by name by using the `data_group_name` parameter.
        For more information on data groups see `Data Groups <https://xradio.readthedocs.io/en/latest/measurement_set_overview.html#Data-Groups>`__ section. See `xarray.Dataset.sel <https://xarray.pydata.org/en/stable/generated/xarray.Dataset.sel.html>`__ for parameter descriptions.

        Returns:
            MeasurementSetXds

        Examples
        --------
        >>> # Select data group 'corrected' and polarization 'XX'.
        >>> selected_ms_xds = ms_xds.sel(data_group_name='corrected', polarization='XX')

        >>> # Select data group 'corrected' and polarization 'XX' using a dict.
        >>> selected_ms_xds = ms_xds.sel({'data_group_name':'corrected', 'polarization':'XX')
        """

        if "data_group_name" in indexers_kwargs:
            data_group_name = indexers_kwargs["data_group_name"]
            del indexers_kwargs["data_group_name"]
        elif (indexers is not None) and ("data_group_name" in indexers):
            data_group_name = indexers["data_group_name"]
            del indexers["data_group_name"]
        else:
            data_group_name = None

        if data_group_name is not None:
            sel_data_group_set = set(
                self.attrs["data_groups"][data_group_name].values()
            )

            data_variables_to_drop = []
            for dg in self.attrs["data_groups"].values():
                temp_set = set(dg.values()) - sel_data_group_set
                data_variables_to_drop.extend(list(temp_set))

            data_variables_to_drop = list(set(data_variables_to_drop))

            sel_ms_xds = MeasurementSetXds(
                super()
                .sel(indexers, method, tolerance, drop, **indexers_kwargs)
                .drop_vars(data_variables_to_drop)
            )

            sel_ms_xds.attrs["data_groups"] = {
                data_group_name: self.attrs["data_groups"][data_group_name]
            }

            return sel_ms_xds
        else:
            return MeasurementSetXds(
                super().sel(indexers, method, tolerance, drop, **indexers_kwargs)
            )



import toolviper.utils.logger as logger
import os
from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np
import xarray as xr


def read_part_keys(inpath: str) -> List[Tuple]:
    """
    Reads the partition keys from a Zarr-stored cds.

    Parameters
    ----------
    inpath : str
        path to read from

    Returns
    -------
    List[Tuple]
        partition keys from a cds

    """

    xds_keys = xr.open_zarr(
        os.path.join(inpath, "partition_keys"),
    )

    spw_ids = xds_keys.coords["spw_ids"]
    pol_setup_ids = xds_keys.coords["pol_setup_ids"]
    intents = xds_keys.coords["intents"]

    return list(zip(spw_ids.values, pol_setup_ids.values, intents.values))


def read_subtables(inpath: str, asdm_subtables: bool) -> Dict[str, xr.Dataset]:
    """
    Reads the metainfo subtables from a Zarr-stored cds.

    Parameters
    ----------
    inpath : str
        path to read from

    asdm_subtables : bool


    Returns
    -------
    Dict[str, xr.Dataset]
        metainfo subtables from a cds

    """

    metainfo = {}
    metadir = Path(inpath, "metainfo")
    for subt in sorted(metadir.iterdir()):
        if subt.is_dir():
            if not asdm_subtables and subt.name.startswith("ASDM_"):
                logger.debug(f"Not loading ASDM_ subtable {subt.name}...")
                continue

            metainfo[subt.name] = read_xds(subt, consolidated=True)

    return metainfo


def read_partitions(inpath: str, part_keys: List[Tuple]) -> Dict[str, xr.Dataset]:
    """
    Reads all the data partitions a Zarr-stored cds.

    Parameters
    ----------
    inpath : str
        path to read from
    part_keys : List[Tuple]


    Returns
    -------
    Dict[str, xr.Dataset]
        partitions from a cds

    """

    partitions = {}
    partdir = Path(inpath, "partitions")
    xds_cnt = 0
    for part in sorted(partdir.iterdir()):
        if part.is_dir() and part.name.startswith("xds_"):
            xds = read_xds(part, consolidated=True)
            partitions[part_keys[xds_cnt]] = xds
            xds_cnt += 1

    return partitions


def read_xds(
    inpath: str,
    chunks: Union[Dict, None] = None,
    consolidated: bool = True,
    overwrite_encoded_chunks: bool = True,
) -> xr.Dataset:
    """
    Read single xds from zarr storage.

    Parameters
    ----------
    inpath : str
        path to read from
    chunks : Union[Dict, None] (Default value = None)
        set chunk size per dimension. Dict is in the form of
        'dim':chunk_size, for example {'time':100, 'baseline':400, 'chan':32, 'pol':1}.
        Default None uses the original chunking in the zarr input.
    consolidated : boold (Default value = True)
        use zarr consolidated metadata.
    overwrite_encoded_chunks : bool (Default value = True)
        drop the zarr chunks encoded for each variable
        when a dataset is loaded with specified chunk sizes.

    Returns
    -------
    xr.Dataset
    """

    xds = xr.open_zarr(
        os.path.join(inpath),
        chunks=chunks,
        consolidated=consolidated,
        overwrite_encoded_chunks=overwrite_encoded_chunks,
    )

    return xds


def read_zarr(
    infile: str,
    sel_xds: Union[List, str] = None,
    chunks: Dict = None,
    consolidated: bool = True,
    overwrite_encoded_chunks: bool = True,
    **kwargs,
):
    """
    Note: old, initial cngi-io format. To be removed, most likely.
    Read zarr format Visibility data from disk to an ngCASA visibilities dataset
    object consisting of dictionaries of xarray Datasets.

    Parameters
    ----------
    infile : str
        input Visibility filename
    sel_xds : string or list
        Select the ddi to open, for example ['xds0','xds1'] will open the first two ddi. Default None returns everything
    chunks : dict
        sets specified chunk size per dimension. Dict is in the form of
        'dim':chunk_size, for example {'time':100, 'baseline':400, 'chan':32, 'pol':1}.
        Default None uses the original zarr chunking.
    consolidated : bool
        use zarr consolidated metadata capability. Only works for stores that have
        already been consolidated. Default True works with datasets produced by
        convert_ms which automatically consolidates metadata.
    overwrite_encoded_chunks : bool
        drop the zarr chunks encoded for each variable when a dataset is loaded with
        specified chunk sizes.  Default True, only applies when chunks is not None.
    **kwargs :


    Returns
    -------

    """

    if chunks is None:
        chunks = "auto"
        # overwrite_encoded_chunks = False
    # print('overwrite_encoded_chunks',overwrite_encoded_chunks)

    infile = os.path.expanduser(infile)
    if sel_xds is None:
        sel_xds = os.listdir(infile)
    sel_xds = list(np.atleast_1d(sel_xds))

    # print(os.path.join(infile, 'DDI_INDEX'))
    mxds = xr.open_zarr(
        os.path.join(infile, "DDI_INDEX"),
        chunks=chunks,
        consolidated=consolidated,
        overwrite_encoded_chunks=overwrite_encoded_chunks,
    )

    for part in os.listdir(os.path.join(infile, "global")):
        xds_temp = xr.open_zarr(
            os.path.join(infile, "global/" + part),
            chunks=chunks,
            consolidated=consolidated,
            overwrite_encoded_chunks=overwrite_encoded_chunks,
        )
        xds_temp = _fix_dict_for_ms(part, xds_temp)
        mxds.attrs[part] = xds_temp.compute()

    for part in os.listdir(infile):
        if ("xds" in part) and (part in sel_xds):
            xds_temp = xr.open_zarr(
                os.path.join(infile, part),
                chunks=chunks,
                consolidated=consolidated,
                overwrite_encoded_chunks=overwrite_encoded_chunks,
            )
            xds_temp = _fix_dict_for_ms(part, xds_temp)
            mxds.attrs[part] = xds_temp

    return mxds


def _fix_dict_for_ms(name, xds):
    # Used to be:
    # xds.attrs["column_descriptions"] = xds.attrs["column_descriptions"][0]
    # xds.attrs["info"] = xds.attrs["info"][0]

    if "xds" in name:
        xds.column_descriptions["UVW"]["shape"] = np.array(
            xds.column_descriptions["UVW"]["shape"].split(",")
        ).astype(int)

    if "spectral_window" == name:
        xds.column_descriptions["CHAN_FREQ"]["keywords"]["MEASINFO"]["TabRefCodes"] = (
            np.array(
                xds.column_descriptions["CHAN_FREQ"]["keywords"]["MEASINFO"][
                    "TabRefCodes"
                ].split(",")
            ).astype(int)
        )
        xds.column_descriptions["REF_FREQUENCY"]["keywords"]["MEASINFO"][
            "TabRefCodes"
        ] = np.array(
            xds.column_descriptions["REF_FREQUENCY"]["keywords"]["MEASINFO"][
                "TabRefCodes"
            ].split(",")
        ).astype(
            int
        )

    if "antenna" == name:
        xds.column_descriptions["OFFSET"]["shape"] = np.array(
            xds.column_descriptions["OFFSET"]["shape"].split(",")
        ).astype(int)
        xds.column_descriptions["POSITION"]["shape"] = np.array(
            xds.column_descriptions["POSITION"]["shape"].split(",")
        ).astype(int)

    if "feed" == name:
        xds.column_descriptions["POSITION"]["shape"] = np.array(
            xds.column_descriptions["POSITION"]["shape"].split(",")
        ).astype(int)

    if "observation" == name:
        xds.column_descriptions["TIME_RANGE"]["shape"] = np.array(
            xds.column_descriptions["TIME_RANGE"]["shape"].split(",")
        ).astype(int)

    return xds



import toolviper.utils.logger as logger
from pathlib import Path
from typing import Dict, Tuple, Union

import dask
import numpy as np
import pandas as pd
import xarray as xr

from casacore import tables

from .table_query import open_query, open_table_ro
from .read import (
    read_col_chunk,
    convert_casacore_time,
    extract_table_attributes,
    add_units_measures,
    table_exists,
    load_generic_table,
)
from .write import revert_time
from xradio._utils.list_and_array import unique_1d


def read_ephemerides(
    infile: str,
) -> Union[xr.Dataset, None]:
    """
    Read ephemerides info from MSv2 FIELD/EPHEMi_....tab subtables

    Parameters
    ----------
    infile : str
        path to MS

    Returns
    -------
    Union[xr.Dataset, None]
        ephemerides xds with metainfo as in the MSv3/EPHEMERIDES subtable
    """
    field_subt = Path(infile, "FIELD")
    subdirs = [
        sdir
        for sdir in field_subt.iterdir()
        if "EPHEM" in sdir.name and sdir.is_dir() and table_exists(str(sdir))
    ]
    ephem = []
    for sdir in subdirs:
        logger.debug(f"Reading ephemerides info from: FIELD / {sdir.name}")
        # One "EPHEM_*.tab" (each with a difference ephemeris_id) to concatenate
        ephem.append(
            load_generic_table(infile, str(Path(*sdir.parts[-2:])), timecols=["MJD"])
        )

    if ephem:
        ephem = xr.concat(ephem, dim="ephemeris_id")
    else:
        ephem = None

    return ephem


def read_delayed_pointing_table(
    infile: str,
    rename_ids: Dict[str, str] = None,
    chunks: Tuple = (10000, 100, 2, 20),
    time_slice=None,
) -> xr.Dataset:
    """
    Read MS pointing subtable in delayed arrays into an xr.Dataset

    Parameters
    ----------
    infile : str
        path to pointing table
    rename_ids : Dict[str, str] (Default value = None)
        dict with dimension renaming mapping
    chunks : Tuple (Default value = (10000, 100, 2, 20))
        chunks for the arrays. Chunks tuple: time, antenna, data_vars_dim_1, data_vars_dim_2
    time_slice: slice
        time bounds

    Returns
    -------
    xr.Dataset
        pointing dataset
    """

    with open_table_ro(infile) as mtable:
        taql_time = ""
        if time_slice:
            times = normalize_time_slice(mtable, time_slice)
            taql_time = f"where TIME BETWEEN {times.start} AND {times.stop}"
        else:
            times = None
        taql_all = f"select * from $mtable {taql_time}"
        with open_query(mtable, taql_all) as query_all:
            if query_all.nrows() == 0:
                mtable.close()
                note = ""
                if taql_time:
                    note = (
                        " within the selected time range: {times.start} - {times.stop}"
                    )
                logger.warning(f"POINTING subtable has no data{note}")
                return xr.Dataset()

            # pointing table uses time x antenna_id
            antennas = unique_1d(query_all.getcol("ANTENNA_ID", 0, -1))
            taql_times = f"select DISTINCT TIME from $mtable {taql_time}"
            with open_query(None, taql_times) as query_times:
                utimes = unique_1d(query_times.getcol("TIME", 0, -1))

            tvars = read_delayed_pointing_times(
                infile, antennas, utimes, chunks, query_all, times
            )

    dims = ["time", "antenna_id", "dim_2", "dim_3"]

    # now concat all the dask chunks from each time
    mvars = {}
    for var in tvars.keys():
        mvars[var] = xr.DataArray(
            dask.array.concatenate(tvars[var], axis=0),
            dims=dims[: len(tvars[var][0].shape)],
        )

    mcoords = {}
    mcoords["time"] = xr.DataArray(convert_casacore_time(utimes), dims=["time"])
    mcoords["antenna_id"] = xr.DataArray(np.arange(len(antennas)), dims=["antenna_id"])

    cc_attrs = extract_table_attributes(infile)
    attrs = {"other": {"msv2": {"ctds_attrs": cc_attrs}}}
    mvars = add_units_measures(mvars, cc_attrs)
    mcoords = add_units_measures(mcoords, cc_attrs)

    xds = xr.Dataset(mvars, coords=mcoords)
    if rename_ids:
        rename_ids = {k: v for k, v in rename_ids.items() if k in xds.sizes}
    xds = xds.rename_dims(rename_ids)
    xds = xds.assign_attrs(attrs)

    return xds


def normalize_time_slice(mtable: tables.table, time_slice: slice) -> slice:
    """
    If we get indices, produce the TIME column time value for the
    start/top indices. If we get timestamps, convert them to casacore
    refeference.

    Parameters
    ----------
    mtable : tables.table
        a casacore table from which we are reading a TIME
        column
    time_slice : slice
        slice giving start/stop time. Can be given as
        integer indices or as timestamps (Xarray / pandas reference)

    Returns
    -------
    slice
        a (start, stop) slice with times in casacore ref frame
    """
    if type(time_slice.start) == pd.Timestamp and type(time_slice.stop) == pd.Timestamp:
        # Add tol?
        eps = np.finfo(float).eps
        times = slice(
            revert_time(time_slice.start) - eps, revert_time(time_slice.stop) + eps
        )

    elif (
        int(time_slice.start) == time_slice.start
        and int(time_slice.stop) == time_slice.stop
    ):
        # instead of int cast could be operator.index(time_slice.start)
        taql_utimes = "select DISTINCT TIME from $mtable"
        with open_query(mtable, taql_utimes) as query_utimes:
            utimes = unique_1d(query_utimes.getcol("TIME", 0, -1))
            # add a tol around the time ranges returned by taql
            if len(utimes) < 2:
                tol = 1e-5
            else:
                tol = np.diff(utimes).min() / 4

        nutimes = len(utimes)
        if nutimes == 0:
            times = slice(0, 0)
        else:
            tidxs = slice(
                min(nutimes, int(time_slice.start)),
                min(nutimes, int(time_slice.stop)) - 1,
            )
            times = slice(utimes[tidxs.start] - tol, utimes[tidxs.stop] + tol)

    else:
        raise ValueError(
            f"Invalid time type. Not a timestamp and cannot use as"
            f" index: {time_slice.start} (type: {type(time_slice.start)})"
        )

    return times


def read_delayed_pointing_times(
    infile: str,
    antennas: np.ndarray,
    utimes: np.array,
    chunks: tuple,
    query_all: tables.table,
    time_slice: slice,
) -> Dict[str, xr.DataArray]:
    """
    Read pointing table in delayed time / antenna chunks. Loops over
    time chunks

    Parameters
    ----------
    infile : str
        path to pointing table
    antennas : np.ndarray
        antenna ids
    utimes : np.ndarray
        unique times from table
    chunks : tuple
        chunks for the arrays
    query_all : tables.table
        table to read columns
    time_slice: slice :
        time bounds

    Returns
    -------
    Dict[str, xr.DataArray]
        dictionary of columns=>variables (read as dask.delayed)
    """

    antenna_chunks = range(0, len(antennas), chunks[1])

    # loop over time chunks
    if time_slice:
        time_chunks = [0]
        logger.debug(
            f"reading single chunk from pointing, with times {time_slice.start} - {time_slice.stop}"
        )
    else:
        time_chunks = range(0, len(utimes), chunks[0])
        logger.debug(
            f"reading pointing table into {len(time_chunks)} time x {len(antenna_chunks)} antenna chunks"
        )

    tvars = {}
    for tc in time_chunks:
        bvars = read_delayed_pointing_chunks(
            infile, antennas, chunks, utimes, tc, query_all
        )

        # now concat all the dask chunks from each antenna
        for var in bvars.keys():
            if len(bvars[var]) == 0:
                continue
            if var not in tvars:
                tvars[var] = []
            tvars[var] += [dask.array.concatenate(bvars[var], axis=1)]

    return tvars


def read_delayed_pointing_chunks(
    infile: str,
    antennas: np.ndarray,
    chunks: tuple,
    utimes: np.ndarray,
    tc: int,
    tb_tool: tables.table,
) -> Dict[str, xr.DataArray]:
    """
    For one time chunk, read the baseline/antenna chunks. Loops over
    antenna_id chunks and reads all columns as dask.delayed calls.

    Parameters
    ----------
    infile : str
        path to pointing table
    antennas : np.ndarray
        antenna ids
    chunks : tuple
        chunks for the arrays
    utimes : np.ndarray
        unique times from table
    tc : int
        time index
    tb_tool : tables.table
        table to read columns

    Returns
    -------
    Dict[str, xr.DataArray]
        dictionary of columns=>variables (read as dask.delayed)
    """

    # add a tol around the time ranges returned by taql, for the next taql queries
    if len(utimes) < 2:
        tol = 1e-5
    else:
        tol = np.diff(utimes).min() / 4

    times = (
        utimes[tc] - tol,
        utimes[min(len(utimes) - 1, tc + chunks[0] - 1)] + tol,
    )
    ctlen = min(len(utimes), tc + chunks[0]) - tc  # chunk time length

    antenna_chunks = range(0, len(antennas), chunks[1])
    cols = tb_tool.colnames()

    bvars = {}
    for bc in antenna_chunks:
        blines = (
            antennas[bc],
            antennas[min(len(antennas) - 1, bc + chunks[1] - 1)],
        )
        cblen = min(len(antennas) - bc, chunks[1])

        # read the specified chunk of data
        ttql = "TIME BETWEEN %f and %f" % times
        atql = "ANTENNA_ID BETWEEN %i and %i" % blines
        ts_taql = f"select * from $mtable where {ttql} AND {atql}"
        with open_query(None, ts_taql) as ts_tb:
            tidxs = np.searchsorted(utimes, ts_tb.getcol("TIME", 0, -1)) - tc
            bidxs = np.searchsorted(antennas, ts_tb.getcol("ANTENNA_ID", 0, -1)) - bc
            didxs = np.arange(len(bidxs))

            # loop over each column and create delayed dask arrays
            for col in cols:
                if (col in ["TIME", "ANTENNA_ID"]) or (
                    not tb_tool.iscelldefined(col, 0)
                ):
                    continue
                if col not in bvars:
                    bvars[col] = []

                cdata = tb_tool.getcol(col, 0, 1)[0]
                if isinstance(cdata, str):
                    cdata = np.array(cdata)
                if len(cdata.shape) == 0:
                    delayed_array = dask.delayed(read_col_chunk)(
                        infile,
                        ts_taql,
                        col,
                        (ctlen, cblen),
                        tidxs,
                        bidxs,
                        didxs,
                        None,
                        None,
                    )
                    bvars[col] += [
                        dask.array.from_delayed(
                            delayed_array, (ctlen, cblen), cdata.dtype
                        )
                    ]

                elif len(cdata.shape) == 2:
                    d1_list = []
                    for cc in range(0, cdata.shape[0], chunks[2]):
                        d1s = (cc, min(cdata.shape[0], cc + chunks[2]) - 1)
                        d2_list = []
                        for pc in range(0, cdata.shape[1], chunks[3]):
                            d2s = (pc, min(cdata.shape[1], pc + chunks[3]) - 1)
                            cshape = (
                                ctlen,
                                cblen,
                            ) + (d1s[1] - d1s[0] + 1, d2s[1] - d2s[0] + 1)
                            delayed_array = dask.delayed(read_col_chunk)(
                                infile,
                                ts_taql,
                                col,
                                cshape,
                                tidxs,
                                bidxs,
                                didxs,
                                d1s,
                                d2s,
                            )
                            d2_list += [
                                dask.array.from_delayed(
                                    delayed_array, cshape, cdata.dtype
                                )
                            ]
                        d1_list += [dask.array.concatenate(d2_list, axis=3)]
                    bvars[col] += [dask.array.concatenate(d1_list, axis=2)]

    return bvars



import copy
import dask.array as da

# import toolviper.utils.logger as logger
import numpy as np
import os
import xarray as xr
import s3fs
from .common import _np_types, _top_level_sub_xds
from ..common import _coords_to_numpy, _dask_arrayize_dv, _numpy_arrayize_dv
from xradio._utils.zarr.common import _get_file_system_and_items


def _read_zarr(
    zarr_store: str, output: dict, selection: dict = {}
) -> (xr.Dataset, bool):
    # supported key/values in output are:
    # "dv"
    #    what data variables should be returned as.
    #    "numpy": numpy arrays
    #    "dask": dask arrays
    # "coords"
    #    what coords should be returned as
    #    "numpy": numpy arrays
    # it's easiest just to copy the object rather than figuring out
    # if it should be copied at some point or not
    # xds = xr.open_zarr(zarr_store).isel(selection).copy(deep=True)
    xds = xr.open_zarr(zarr_store).isel(selection)
    do_dask = False
    do_numpy = False
    do_np_coords = False
    if "dv" in output:
        dv = output["dv"]
        if dv in ["dask", "numpy"]:
            do_dask = dv == "dask"
            do_numpy = not do_dask
        else:
            raise ValueError(
                f"Unsupported value {output[dv]} for output[dv]. "
                "Supported values are 'dask' and 'numpy'"
            )
    if "coords" in output:
        c = output["coords"]
        if c == "numpy":
            do_np_coords = True
        else:
            raise ValueError(
                f"Unexpected value {c} for output[coords]. "
                "The supported value is 'numpy'"
            )
    # do not pass selection, because that is only for the top level data vars
    xds = _decode(xds, zarr_store, output)
    if do_np_coords:
        xds = _coords_to_numpy(xds)
    if do_dask:
        xds = _dask_arrayize_dv(xds)
    elif do_numpy:
        xds = _numpy_arrayize_dv(xds)
    return xds


def _decode(xds: xr.Dataset, zarr_store: str, output: dict) -> (xr.Dataset, bool):
    xds.attrs = _decode_dict(xds.attrs, "")
    sub_xdses = _decode_sub_xdses(zarr_store, output)
    for k, v in sub_xdses.items():
        xds.attrs[k] = v
    return xds


def _decode_dict(my_dict: dict, top_key: str) -> dict:
    # Decodes numpy arrays
    my_dict
    for k, v in my_dict.items():
        if isinstance(v, dict):
            if (
                "_type" in v
                and v["_type"] == "numpy.ndarray"
                and "_value" in v
                and "_dtype" in v
            ):
                my_dict[k] = np.array(v["_value"], dtype=_np_types[v["_dtype"]])
            else:
                z = os.sep.join([top_key, k]) if top_key else k
                my_dict[k] = _decode_dict(v, z)
    return my_dict


def _decode_sub_xdses(zarr_store: str, output: dict) -> dict:
    sub_xdses = {}

    fs, store_contents = _get_file_system_and_items(zarr_store)

    if isinstance(fs, s3fs.core.S3FileSystem):
        # could we just use the items as returned from the helper function..?
        store_tree = fs.walk(zarr_store, topdown=True)
        prepend_s3 = "s3://"
    else:
        store_tree = os.walk(zarr_store, topdown=True)
        prepend_s3 = ""

    for root, dirs, files in store_tree:
        for d in dirs:
            if d.startswith(_top_level_sub_xds):
                xds = _read_zarr(os.sep.join([root, d]), output)
                # for k, v in xds.data_vars.items():
                #    xds = xds.drop_vars([k]).assign({k: v.compute()})
                ky = d[len(_top_level_sub_xds) + 1 :]
                sub_xdses[ky] = xds

    return sub_xdses



import toolviper.utils.logger as logger
import os
import time
from typing import Tuple, Union

import numpy as np
import xarray as xr

from xradio._utils.coord_math import convert_to_si_units
from xradio._utils.schema import (
    column_description_casacore_to_msv4_measure,
    convert_generic_xds_to_xradio_schema,
)
from .subtables import subt_rename_ids
from ._tables.read import (
    load_generic_table,
    make_taql_where_between_min_max,
    table_exists,
    table_has_column,
)


standard_time_coord_attrs = {
    "type": "time",
    "units": ["s"],
    "scale": "utc",
    "format": "unix",
}


def rename_and_interpolate_to_time(
    xds: xr.Dataset,
    time_initial_name: str,
    interp_time: Union[xr.DataArray, None],
    message_prefix: str,
) -> xr.Dataset:
    """
    This function interpolates the time dimension and renames it:

    - interpolates a time_* dimension to values given in interp_time (presumably the time
    axis of the main xds)
    - rename/replace that time_* dimension to "time", where time_* is a (sub)xds specific
    time axis
    (for example "time_pointing", "time_ephemeris", "time_syscal", "time_phase_cal").

    If interp_time is None this will simply return the input xds without modificaitons.
    Uses interpolate_to_time() for interpolation.
    ...

    Parameters:
    ----------
    xds : xr.Dataset
        Xarray dataset to interpolate (presumably a pointing_xds or an xds of
        ephemeris variables)
    time_initial_name: str = None
        Name of time to be renamed+interpolated. Expected an existing time_* coordinate in the
        dataset
    interp_time:
        Time axis to interpolate the dataset to (usually main MSv4 time)
    message_prefix:
        A prefix for info/debug/etc. messages about the specific xds being interpolated/
        time-renamed

    Returns:
    -------
    renamed_interpolated_xds : xr.Dataset
        xarray dataset with time axis renamed to "time" (from time_name, for example
        "time_ephemeris") and interpolated to interp_time.
    """
    if interp_time is None:
        return xds

    interpolated_xds = interpolate_to_time(
        xds,
        interp_time,
        message_prefix,
        time_name=time_initial_name,
    )

    # rename the time_* axis to time.
    time_coord = {"time": (time_initial_name, interp_time.data)}
    renamed_time_xds = interpolated_xds.assign_coords(time_coord)
    renamed_time_xds.coords["time"].attrs.update(standard_time_coord_attrs)
    renamed_time_xds = renamed_time_xds.swap_dims({time_initial_name: "time"})
    if time_initial_name != "time":
        renamed_time_xds = renamed_time_xds.drop_vars(time_initial_name)

    return renamed_time_xds


def interpolate_to_time(
    xds: xr.Dataset,
    interp_time: Union[xr.DataArray, None],
    message_prefix: str,
    time_name: str = "time",
) -> xr.Dataset:
    """
    Interpolate the time coordinate of the input xarray dataset to the
    a data array. This can be used for example to interpolate a pointing_xds
    to the time coord of the (main) MSv4, or similarly the ephemeris
    data variables of a field_and_source_xds.

    Uses interpolation method "linear", unless the source number of points is
    1 in which case "nearest" is used, to avoid divide-by-zero issues.

    Parameters:
    ----------
    xds : xr.Dataset
        Xarray dataset to interpolate (presumably a pointing_xds or an xds of
        ephemeris variables)
    interp_time : Union[xr.DataArray, None]
        Time axis to interpolate the dataset to (usually main MSv4 time)
    message_prefix: str
        A prefix for info/debug/etc. messages

    Returns:
    -------
    interpolated_xds : xr.Dataset
        xarray dataset with time axis interpolated to interp_time.
    """
    if interp_time is not None:
        points_before = xds[time_name].size
        if points_before > 1:
            method = "linear"
        else:
            method = "nearest"
        xds = xds.interp(
            {time_name: interp_time.data}, method=method, assume_sorted=True
        )
        # scan_number sneaks in as a coordinate of the main time axis, drop it
        if "scan_number" in xds.coords:
            xds = xds.drop_vars("scan_number")
        points_after = xds[time_name].size
        logger.debug(
            f"{message_prefix}: interpolating the time coordinate "
            f"from {points_before} to {points_after} points"
        )

    return xds


def make_taql_where_weather(
    in_file: str, ant_xds_station_name_ids: xr.DataArray
) -> str:
    """
    The use of taql_where with WEATHER is complicated because of ALMA and its
    (ab)use of NS_WX_STATION_ID vs. ANTENNA_ID (all -1) (see read.py). We
    cannot simply use a 'WHERE ANTENNA_ID=...'. This function produces a TaQL
    where string that uses ANTENNA_ID/NS_WX_STATION_ID depending on waht is
    found in the WEATHER subtable.

    Parameters
    ----------
    in_file : str
        Input MS name.
    ant_xds_station_name_ids : xr.DataArray
        station name data array from antenna_xds, with name/id information

    Returns
    -------
    taql_where: str
        WHERE substring safe to use in taql_where when loading WEATHER subtable
    """
    weather_path = os.path.join(in_file, "WEATHER")
    if table_exists(weather_path) and table_has_column(
        weather_path, "NS_WX_STATION_ID"
    ):
        taql_where = f"WHERE (NS_WX_STATION_ID IN [{','.join(map(str, ant_xds_station_name_ids.antenna_id.values))}])"
    else:
        taql_where = f"WHERE (ANTENNA_ID IN [{','.join(map(str, ant_xds_station_name_ids.antenna_id.values))}])"

    return taql_where


def create_weather_xds(in_file: str, ant_xds_station_name_ids: xr.DataArray):
    """
    Creates a Weather Xarray Dataset from a MS v2 WEATHER table.

    Parameters
    ----------
    in_file : str
        Input MS name.
    ant_xds_station_name_ids : xr.DataArray
        station name data array from antenna_xds, with name/id information

    Returns
    -------
    xr.Dataset
        Weather Xarray Dataset.
    """

    try:
        taql_where = make_taql_where_weather(in_file, ant_xds_station_name_ids)
        generic_weather_xds = load_generic_table(
            in_file,
            "WEATHER",
            rename_ids=subt_rename_ids["WEATHER"],
            taql_where=taql_where,
        )
    except ValueError as _exc:
        return None

    if not generic_weather_xds.data_vars:
        # for example when the weather subtable only has info for antennas/stations
        # not present in the MSv4 (no overlap between antennas loaded in ant_xds and weather)
        return None

    weather_xds = xr.Dataset(attrs={"type": "weather"})
    stations_present = ant_xds_station_name_ids.sel(
        antenna_id=generic_weather_xds["ANTENNA_ID"]
    )
    coords = {
        "station_name": stations_present.data,
        "antenna_name": stations_present.coords["antenna_name"].data,
    }
    weather_xds = weather_xds.assign_coords(coords)

    dims_station_time = ["station_name", "time_weather"]
    to_new_data_variables = {
        "H20": ["H2O", dims_station_time],
        "IONOS_ELECTRON": ["IONOS_ELECTRON", dims_station_time],
        "PRESSURE": ["PRESSURE", dims_station_time],
        "REL_HUMIDITY": ["REL_HUMIDITY", dims_station_time],
        "TEMPERATURE": ["TEMPERATURE", dims_station_time],
        "DEW_POINT": ["DEW_POINT", dims_station_time],
        "WIND_DIRECTION": ["WIND_DIRECTION", dims_station_time],
        "WIND_SPEED": ["WIND_SPEED", dims_station_time],
    }

    to_new_coords = {
        "TIME": ["time_weather", ["time_weather"]],
    }

    weather_xds = convert_generic_xds_to_xradio_schema(
        generic_weather_xds, weather_xds, to_new_data_variables, to_new_coords
    )

    # TODO: option to interpolate to main time

    # PRESSURE: hPa in MSv2 specs and some MSs => Pa
    weather_xds = convert_to_si_units(weather_xds)

    # correct expected types (for example "IONOS_ELECTRON", "PRESSURE" can be float32)
    for data_var in weather_xds:
        if weather_xds.data_vars[data_var].dtype != np.float64:
            weather_xds[data_var] = weather_xds[data_var].astype(np.float64)

    return weather_xds


def correct_generic_pointing_xds(
    generic_pointing_xds: xr.Dataset, to_new_data_variables: dict[str, list]
) -> xr.Dataset:
    """
    Takes a (generic) pointing_xds as read from a POINTING subtable of an MSv2
    and tries to correct several deviations from the MSv2 specs seen in
    common test data.
    The problems fixed here include wrong dimensions:
    - for example transposed dimensions with respect to the MSv2 specs (output
    from CASA simulator),
    - missing/additional unexpected dimensions when some of the columns are
    empty (in the sense of "empty casacore cells").

    This function modifies the data arrays of the data vars affected by such
    issues.

    Parameters
    ----------
    generic_pointing_xds: xr.Dataset
        The generic pointing dataset (loaded from MSv2) to be fixed
    to_new_data_variables: dict
        The dict used for convert_generic_xds_to_xradio_schema, which gives all
        the data variables relevant for the final MSv4 dataset.

    Returns:
    --------
    xr.Dataset
        Corrected dataset with dimensions conforming to MSv2 specs.
    """

    correct_pointing_xds = generic_pointing_xds.copy()

    for data_var_name in generic_pointing_xds:
        if data_var_name in to_new_data_variables:
            # Corrects dim sizes of "empty cell" variables, such as empty DIRECTION, TARGET, etc.
            if (
                "dim_2" in generic_pointing_xds.sizes
                and generic_pointing_xds.sizes["dim_2"] == 0
            ):
                # When all direction variables are "empty"
                data_var_data = xr.DataArray(
                    [[[[np.nan, np.nan]]]],
                    dims=generic_pointing_xds.dims,
                ).isel(n_polynomial=0, drop=True)
                correct_pointing_xds[data_var_name].data = data_var_data

            elif (
                "dir" in generic_pointing_xds.sizes
                and generic_pointing_xds.sizes["dir"] == 0
            ):
                # When some direction variables are "empty" but some are populated properly
                if "dim_2" in generic_pointing_xds[key].sizes:
                    data_var_data = xr.DataArray(
                        generic_pointing_xds[key].values,
                        dims=generic_pointing_xds[key].dims,
                    )
                else:
                    shape = tuple(
                        generic_pointing_xds.sizes[dim]
                        for dim in ["TIME", "ANTENNA_ID"]
                    ) + (2,)
                    data_var_data = xr.DataArray(
                        np.full(shape, np.nan),
                        dims=generic_pointing_xds[key].dims,
                    )
                correct_pointing_xds[data_var_name].data = data_var_data

    return correct_pointing_xds


def create_pointing_xds(
    in_file: str,
    ant_xds_name_ids: xr.DataArray,
    time_min_max: Union[Tuple[np.float64, np.float64], None],
    interp_time: Union[xr.DataArray, None] = None,
) -> xr.Dataset:
    """
    Creates a Pointing Xarray Dataset from an MS v2 POINTING (sub)table.

    WIP: details of a few direction variables (and possibly moving some to attributes) to be
    settled (see MSv4 spreadsheet).

    Parameters
    ----------
    in_file : str
        Input MS name.
    ant_xds_name_ids : xr.DataArray
        antenna_name data array from antenna_xds, with name/id information
    time_min_max : tuple
        min / max times values to constrain loading (from the TIME column)
    interp_time : Union[xr.DataArray, None] (Default value = None)
        interpolate time to this (presumably main dataset time)

    Returns
    -------
    xr.Dataset
         Pointing Xarray dataset
    """
    start = time.time()

    taql_time_range = make_taql_where_between_min_max(
        time_min_max, in_file, "POINTING", "TIME"
    )

    if taql_time_range is None:
        taql_where = f"WHERE (ANTENNA_ID IN [{','.join(map(str, ant_xds_name_ids.antenna_id.values))}])"
    else:
        taql_where = (
            taql_time_range
            + f" AND (ANTENNA_ID IN [{','.join(map(str, ant_xds_name_ids.antenna_id.values))}])"
        )
    # Read POINTING table into a Xarray Dataset.
    generic_pointing_xds = load_generic_table(
        in_file,
        "POINTING",
        rename_ids=subt_rename_ids["POINTING"],
        taql_where=taql_where,
    )

    if not generic_pointing_xds.data_vars:
        # apparently empty MS/POINTING table => produce empty xds
        return xr.Dataset()

    # Checking a simple way of using only the one single coefficient of the polynomials
    if "n_polynomial" in generic_pointing_xds.sizes:
        size = generic_pointing_xds.sizes["n_polynomial"]
        if size == 1:
            generic_pointing_xds = generic_pointing_xds.sel({"n_polynomial": 0})
        elif size == 0:
            generic_pointing_xds = generic_pointing_xds.drop_dims("n_polynomial")

    time_ant_dims = ["time_pointing", "antenna_name"]
    time_ant_dir_dims = time_ant_dims + ["local_sky_dir_label"]
    to_new_data_variables = {
        "DIRECTION": ["POINTING_BEAM", time_ant_dir_dims],
        "ENCODER": ["POINTING_DISH_MEASURED", time_ant_dir_dims],
        "OVER_THE_TOP": ["POINTING_OVER_THE_TOP", time_ant_dims],
    }

    to_new_coords = {
        "TIME": ["time_pointing", ["time_pointing"]],
        "dim_2": ["local_sky_dir_label", ["local_sky_dir_label"]],
    }

    generic_pointing_xds = correct_generic_pointing_xds(
        generic_pointing_xds, to_new_data_variables
    )
    pointing_xds = xr.Dataset(attrs={"type": "pointing"})
    coords = {
        "antenna_name": ant_xds_name_ids.sel(
            antenna_id=generic_pointing_xds["ANTENNA_ID"]
        ).data,
        "local_sky_dir_label": ["az", "alt"],
    }
    pointing_xds = pointing_xds.assign_coords(coords)
    pointing_xds = convert_generic_xds_to_xradio_schema(
        generic_pointing_xds, pointing_xds, to_new_data_variables, to_new_coords
    )

    pointing_xds = rename_and_interpolate_to_time(
        pointing_xds, "time_pointing", interp_time, "pointing_xds"
    )

    logger.debug(f"create_pointing_xds() execution time {time.time() - start:0.2f} s")

    return pointing_xds


def prepare_generic_sys_cal_xds(generic_sys_cal_xds: xr.Dataset) -> xr.Dataset:
    """
    A generic_sys_cal_xds loaded with load_generic_table() cannot be easily
    used in convert_generic_xds_to_xradio_schema() to produce an MSv4
    sys_cal_xds dataset, as their structure differs in dimensions and order
    of dimensions.
    This function performs various prepareation steps, such as:
    - filter out dimensions not neeed for an individual MSv4 (SPW, FEED),
    - drop variables loaded from columns with all items set to empty array,
    - transpose the dimensions frequency,receptor
    - fix dimension names (and order) when needed.

    Parameters
    ----------
    generic_sys_cal_xds : xr.Dataset
        generic dataset read from an MSv2 SYSCAL subtable

    Returns
    -------
    generic_sys_cal_xds: xr.Dataset
        System calibration Xarray Dataset prepared for generic conversion
        to MSv4.
    """

    # drop SPW and feed dims
    generic_sys_cal_xds = generic_sys_cal_xds.isel(SPECTRAL_WINDOW_ID=0, drop=True)
    generic_sys_cal_xds = generic_sys_cal_xds.isel(FEED_ID=0, drop=True)

    # Often some of the T*_SPECTRUM are present but all the cells are populated
    # with empty arrays
    empty_arrays_vars = []
    for data_var in generic_sys_cal_xds.data_vars:
        if generic_sys_cal_xds[data_var].size == 0:
            empty_arrays_vars.append(data_var)
    if empty_arrays_vars:
        generic_sys_cal_xds = generic_sys_cal_xds.drop_vars(empty_arrays_vars)

    # Re-arrange receptor and frequency dims depending on input structure
    if (
        "receptor" in generic_sys_cal_xds.sizes
        and "frequency" in generic_sys_cal_xds.sizes
    ):
        # dim_3 can be created for example when the T*_SPECTRUM have varying # channels!
        # more generaly, could transpose with ... to avoid errors with additional spurious dimensions
        if "dim_3" in generic_sys_cal_xds.dims:
            generic_sys_cal_xds = generic_sys_cal_xds.drop_dims("dim_3")
        # From MSv2 tables we get (...,frequency, receptor)
        #  -> transpose to (...,receptor,frequency) ready for MSv4 sys_cal_xds
        generic_sys_cal_xds = generic_sys_cal_xds.transpose(
            "ANTENNA_ID", "TIME", "receptor", "frequency"
        )
    elif (
        "frequency" in generic_sys_cal_xds.sizes
        and not "dim_3" in generic_sys_cal_xds.sizes
    ):
        # because order is (...,frequency,receptor), when frequency is missing
        # receptor can get wrongly labeled as frequency
        generic_sys_cal_xds = generic_sys_cal_xds.rename_dims({"frequency": "receptor"})
    elif (
        "frequency" not in generic_sys_cal_xds.sizes
        and "receptor" in generic_sys_cal_xds.sizes
        and "dim_3" in generic_sys_cal_xds.sizes
    ):
        # different *_SPECTRUM array sizes + some empty arrays can create an additional spurious
        # generic dimension, which should have been "receptor"
        generic_sys_cal_xds = generic_sys_cal_xds.rename_dims({"receptor": "frequency"})
        generic_sys_cal_xds = generic_sys_cal_xds.rename_dims({"dim_3": "receptor"})
        generic_sys_cal_xds = generic_sys_cal_xds.transpose(
            "ANTENNA_ID", "TIME", "receptor", "frequency"
        )
    else:
        raise RuntimeError(
            "Cannot understand the arrangement of dimensions of {generic_sys_cal_xds=}"
        )

    return generic_sys_cal_xds


def create_system_calibration_xds(
    in_file: str,
    main_xds_frequency: xr.DataArray,
    ant_xds_name_ids: xr.DataArray,
    sys_cal_interp_time: Union[xr.DataArray, None] = None,
):
    """
    Creates a system calibration Xarray Dataset from a MSv2 SYSCAL table.

    Parameters
    ----------
    in_file: str
        Input MS name.
    main_xds_frequency: xr.DataArray
        frequency array of the main xds (MSv4), containing among other things
        spectral_window_id and measures metadata
    ant_xds_name_ids : xr.Dataset
        antenna_name data array from antenna_xds, with name/id information
    sys_cal_interp_time: Union[xr.DataArray, None] = None,
        Time axis to interpolate the data vars to (usually main MSv4 time)

    Returns
    -------
    sys_cal_xds: xr.Dataset
        System calibration Xarray Dataset.
    """

    spectral_window_id = main_xds_frequency.attrs["spectral_window_id"]
    try:
        generic_sys_cal_xds = load_generic_table(
            in_file,
            "SYSCAL",
            rename_ids=subt_rename_ids["SYSCAL"],
            taql_where=(
                f" where (SPECTRAL_WINDOW_ID = {spectral_window_id})"
                f" AND (ANTENNA_ID IN [{','.join(map(str, ant_xds_name_ids.antenna_id.values))}])"
            ),
        )
    except ValueError as _exc:
        return None

    if not generic_sys_cal_xds.data_vars:
        # even though SYSCAL is an optional subtable, some write it empty
        return None

    generic_sys_cal_xds = prepare_generic_sys_cal_xds(generic_sys_cal_xds)

    mandatory_dimensions = ["antenna_name", "time_cal", "receptor_label"]
    if "frequency" not in generic_sys_cal_xds.sizes:
        dims_all = mandatory_dimensions
    else:
        dims_all = mandatory_dimensions + ["frequency_cal"]

    to_new_data_variables = {
        "PHASE_DIFF": ["PHASE_DIFFERENCE", ["antenna_name", "time_cal"]],
        "TCAL": ["TCAL", dims_all],
        "TCAL_SPECTRUM": ["TCAL", dims_all],
        "TRX": ["TRX", dims_all],
        "TRX_SPECTRUM": ["TRX", dims_all],
        "TSKY": ["TSKY", dims_all],
        "TSKY_SPECTRUM": ["TSKY", dims_all],
        "TSYS": ["TSYS", dims_all],
        "TSYS_SPECTRUM": ["TSYS", dims_all],
        "TANT": ["TANT", dims_all],
        "TANT_SPECTRUM": ["TANT", dims_all],
        "TAN_TSYS": ["TAN_TSYS", dims_all],
        "TANT_SYS_SPECTRUM": ["TANT_TSYS", dims_all],
    }

    to_new_coords = {
        "TIME": ["time_cal", ["time_cal"]],
        "receptor": ["receptor_label", ["receptor_label"]],
        "frequency": ["frequency_cal", ["frequency_cal"]],
    }

    sys_cal_xds = xr.Dataset(attrs={"type": "system_calibration"})
    coords = {
        "antenna_name": ant_xds_name_ids.sel(
            antenna_id=generic_sys_cal_xds["ANTENNA_ID"]
        ).data,
        "receptor_label": generic_sys_cal_xds.coords["receptor"].data,
    }
    sys_cal_xds = sys_cal_xds.assign_coords(coords)
    sys_cal_xds = convert_generic_xds_to_xradio_schema(
        generic_sys_cal_xds, sys_cal_xds, to_new_data_variables, to_new_coords
    )

    # Add frequency coord and its measures data, if present
    if "frequency_cal" in dims_all:
        frequency_coord = {
            "frequency_cal": generic_sys_cal_xds.coords["frequency"].data
        }
        sys_cal_xds = sys_cal_xds.assign_coords(frequency_coord)
        frequency_measure = {
            "type": main_xds_frequency.attrs["type"],
            "units": main_xds_frequency.attrs["units"],
            "observer": main_xds_frequency.attrs["observer"],
        }
        sys_cal_xds.coords["frequency_cal"].attrs.update(frequency_measure)

    sys_cal_xds = rename_and_interpolate_to_time(
        sys_cal_xds, "time_cal", sys_cal_interp_time, "system_calibration_xds"
    )

    # correct expected types
    for data_var in sys_cal_xds:
        if sys_cal_xds.data_vars[data_var].dtype != np.float64:
            sys_cal_xds[data_var] = sys_cal_xds[data_var].astype(np.float64)

    return sys_cal_xds



import importlib
import dataclasses
import typing

from docutils import nodes, utils
from docutils.parsers.rst import Directive, DirectiveError
from docutils.parsers.rst import directives
from docutils.utils import SystemMessagePropagation
from docutils.statemachine import StringList

from sphinx.directives import ObjectDescription
from sphinx.util.docutils import switch_source_input

from xradio.schema import (
    xarray_dataclass_to_array_schema,
    xarray_dataclass_to_dataset_schema,
    xarray_dataclass_to_dict_schema,
)


class SchemaTableDirective(ObjectDescription):
    required_arguments = 1  # Xarray dataclass of schema
    has_content = False

    option_spec = {
        # "headers": directives.unchanged,
        # "widths": directives.unchanged,
        "title": directives.unchanged,
        # "columns": directives.unchanged,
        "class": directives.unchanged,
    }

    def run(self):
        # Import the referenced class
        klass_path = self.arguments[0].rsplit(".", 1)
        if len(klass_path) != 2:
            raise ValueError(
                f"Should be absolute Python name of xarray dataclass definition: {self.arguments[0]}"
            )
        klass_module = importlib.import_module(klass_path[0])
        klass = getattr(klass_module, klass_path[1])

        # Make table node
        classes = ["tbl", "colwidths-given"]
        if "class" in self.options:
            classes.append(self.options["class"])
        self._table = nodes.table(
            classes=classes, ids=[f"schema-table-{self.arguments[0]}"]
        )

        # Add title, if requested
        if "title" in self.options:
            self._table += nodes.title(text=caption)

        # Declare columns
        column_widths = [10, 10, 5, 5, 40]
        self._tgroup = nodes.tgroup(cols=len(column_widths))
        self._table += self._tgroup
        for colwidth in column_widths:
            self._tgroup += nodes.colspec(colwidth=colwidth)

        # Create head row
        header_row = nodes.row()
        header_row += nodes.entry("", nodes.paragraph(text=""))
        header_row += nodes.entry("", nodes.strong(text="Dimensions"))
        header_row += nodes.entry("", nodes.strong(text="Dtype"))
        header_row += nodes.entry("", nodes.strong(text="Model"))
        header_row += nodes.entry("", nodes.strong(text="Description"))
        self._thead = nodes.thead("", header_row)
        self._tgroup += self._thead

        # Add body
        self._tbody = nodes.tbody()
        self._tgroup += self._tbody

        # Add table contents (overridden in subclasses)
        self._add_table_contents(klass)

        # Register table
        # tbl = self.env.get_domain("tbl")
        # tbl.add_table(caption, table_id)

        return [self._table]

    def _add_section(self, name):
        # Create row
        row = nodes.row("", nodes.entry("", nodes.strong("", name), morecols=4))
        self._tbody += row

    def _add_row(
        self,
        name="",
        dimss=[],
        types=[],
        meta=None,
        descr="",
        optional=False,
        default=dataclasses.MISSING,
    ):
        # Create row
        row = nodes.row()
        self._tbody += row

        # Add name
        name_nds = [nodes.literal(text=name)]
        if optional:
            name_nds = [nodes.Text("(")] + name_nds + [nodes.Text(")")]
        row += nodes.entry("", *name_nds)

        # Add dimensions
        def mk_multi_entry(lines):
            if not lines:
                return nodes.entry()
            if len(lines) == 0:
                return nodes.entry(lines[0])
            else:
                entry = nodes.entry()
                for i, line in enumerate(lines):
                    entry += nodes.line("", "" if i == 0 else "or\xa0", line)
                return entry

        row += mk_multi_entry(
            [nodes.literal(text=f"[{','.join(dims)}]") for dims in dimss]
        )

        # Add types
        row += mk_multi_entry([nodes.literal(text=typ) for typ in types])

        # Add model link
        entry = nodes.entry()
        row += entry
        if meta is not None:
            # Preformatted? Just pass through
            if isinstance(meta, nodes.line):
                entry += meta
            else:
                vl = StringList()
                vl.append(f":py:class:`~{meta}`", "")
                with switch_source_input(self.state, vl):
                    self.state.nested_parse(vl, 0, entry)

        # Add description
        entry = nodes.entry()
        row += entry
        if descr:
            vl = StringList()
            vl.append(descr, "")
            with switch_source_input(self.state, vl):
                self.state.nested_parse(vl, 0, entry)
        if default is not dataclasses.MISSING:
            vl = StringList()
            vl.append(f"**Default:** ``{repr(default)}``", "")
            with switch_source_input(self.state, vl):
                self.state.nested_parse(vl, 0, entry)


def format_literals(typ):

    # a | b | c: Recurse and merge
    if typing.get_origin(typ) == typing.Union:
        type_args = typing.get_args(typ)
        options = []
        for arg in type_args:
            options += format_literals(arg)
        return options

    # Literal['a', 'b', ...]: Wrap into individual "literal" nodes
    if typing.get_origin(typ) == typing.Literal:
        return list(map(lambda t: nodes.literal(text=repr(t)), typing.get_args(typ)))

    # list[Literal['a'], Literal['b'], ...]: Format as one literal (compound) value
    if typing.get_origin(typ) == list:
        type_args = typing.get_args(typ)
        if any([typing.get_origin(arg) != typing.Literal for arg in type_args]):
            raise ValueError(f"List must contain only literals: {typ}")
        values = [repr(typing.get_args(val)[0]) for val in typing.get_args(typ)]
        return [nodes.literal(text=f"[{', '.join(values)}]")]

    raise ValueError(f"Must be either a type or a literal: {typ}")


def format_attr_model_text(state, attr) -> StringList:
    """
    Formats the text for the 'model' column in schema tables (arrays and datasets).
    Doesn't aim at supporting any literal types or combinations of types in general,
    but the following three ones specifically:

    - Literals (with multiple options (implicit Union of literals))
    - List of literals (e.g. ["rad","rad"]
    - Union of list of literals (e.g. ["m","m","m"]/["rad","rad","m"]

    This is meant to produce readable text listing literals as quoted text and
    their combinations, in schema attributes (particularly quantities and measures).

    Everything else than these expected literal based types would be printed as the
    type name.
    """

    type_args = typing.get_args(attr.typ)
    is_list_of_literals = typing.get_origin(attr.typ) is list and all(
        [typing.get_origin(arg) is typing.Literal for arg in type_args]
    )

    line = nodes.line()

    if not is_list_of_literals:
        # A type?
        if isinstance(attr.typ, type):
            vl = StringList()
            vl.append(f":py:class:`~{attr.typ.__module__}.{attr.typ.__name__}`", "")
            with switch_source_input(state, vl):
                state.nested_parse(vl, 0, line)
            return line

        # Derived type, e.g. list of types?
        if typing.get_origin(attr.typ) == list and all(
            [isinstance(arg, type) for arg in type_args]
        ):
            vl = StringList()
            vl.append("[", "")
            for i, arg in enumerate(typing.get_args(attr.typ)):
                if i > 0:
                    vl.append(", ", "")
                vl.append(f":py:class:`~{arg.__module__}.{arg.__name__}`", "")
            vl.append("]", "")
            with switch_source_input(state, vl):
                state.nested_parse(vl, 0, line)
            return line

    # Assume it's a literal of some kind - collect options
    literals = format_literals(attr.typ)
    for i, lit in enumerate(literals):
        if i > 0:
            if i + 1 >= len(literals):
                line += nodes.Text(" or\xa0")
            else:
                line += nodes.Text(", ")
        line += lit

    return line


class ArraySchemaTableDirective(SchemaTableDirective):
    def _add_table_contents(self, klass):
        # Extract schema
        schema = xarray_dataclass_to_array_schema(klass)

        # Add dataarray reference as first element
        self._add_row(
            "data",
            schema.dimensions,
            schema.dtypes,
            schema.schema_name,
            schema.data_docstring,
        )

        # Add coordinates
        if schema.coordinates:
            self._add_section("Coordinates:")
            for coord in schema.coordinates:
                self._add_row(
                    coord.name,
                    coord.dimensions,
                    coord.dtypes,
                    coord.schema_name,
                    coord.docstring or coord.data_docstring,
                    optional=coord.optional,
                    default=coord.default,
                )

        # Add attributes
        if schema.attributes:
            self._add_section("Attributes:")
            for attr in schema.attributes:
                model_text = format_attr_model_text(self.state, attr)
                self._add_row(
                    attr.name,
                    [],
                    [],
                    model_text,
                    attr.docstring,
                    optional=attr.optional,
                    default=attr.default,
                )


class DatasetSchemaTableDirective(SchemaTableDirective):
    def _add_table_contents(self, klass):
        # Extract schema
        schema = xarray_dataclass_to_dataset_schema(klass)

        # Add coordinates
        if schema.coordinates:
            self._add_section("Coordinates:")
            for coord in schema.coordinates:
                self._add_row(
                    coord.name,
                    coord.dimensions,
                    coord.dtypes,
                    coord.schema_name,
                    coord.docstring or coord.data_docstring,
                    optional=coord.optional,
                    default=coord.default,
                )

        # Add data variables
        if schema.data_vars:
            self._add_section("Data Variables:")
            for data_var in schema.data_vars:
                self._add_row(
                    data_var.name,
                    data_var.dimensions,
                    data_var.dtypes,
                    data_var.schema_name,
                    data_var.docstring or data_var.data_docstring,
                    optional=data_var.optional,
                    default=data_var.default,
                )

        # Add attributes
        if schema.attributes:
            self._add_section("Attributes:")
            for attr in schema.attributes:
                model_text = format_attr_model_text(self.state, attr)
                self._add_row(
                    attr.name,
                    [],
                    [],
                    model_text,
                    attr.docstring,
                    optional=attr.optional,
                    default=attr.default,
                )


class DictSchemaTableDirective(SchemaTableDirective):
    def _add_table_contents(self, klass):
        # Extract schema
        schema = xarray_dataclass_to_dict_schema(klass)

        # Add attributes
        if schema.attributes:
            self._add_section("Fields:")
            for attr in schema.attributes:
                self._add_row(
                    attr.name,
                    types=[f"{attr.typ.__name__}"],
                    optional=attr.optional,
                    descr=attr.docstring,
                    default=attr.default,
                )



from astropy.wcs import WCS
import numpy as np
import xarray as xr
from typing import List, Union
from .common import _c, _compute_world_sph_dims, _l_m_attr_notes
from xradio._utils.coord_math import _deg_to_rad
from xradio._utils.dict_helpers import make_quantity


def _input_checks(
    phase_center: Union[list, np.ndarray],
    image_size: Union[list, np.ndarray],
    cell_size: Union[list, np.ndarray],
) -> None:
    if len(image_size) != 2:
        raise ValueError("image_size must have exactly two elements")
    if len(phase_center) != 2:
        raise ValueError("phase_center must have exactly two elements")
    if len(cell_size) != 2:
        raise ValueError("cell_size must have exactly two elements")


def _make_coords(
    chan_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
) -> dict:
    if not isinstance(chan_coords, list) and not isinstance(chan_coords, np.ndarray):
        chan_coords = [chan_coords]
    chan_coords = np.array(chan_coords, dtype=np.float64)
    restfreq = chan_coords[len(chan_coords) // 2]
    vel_coords = (1 - chan_coords / restfreq) * _c
    if not isinstance(time_coords, list) and not isinstance(time_coords, np.ndarray):
        time_coords = [time_coords]
    time_coords = np.array(time_coords, dtype=np.float64)
    return dict(chan=chan_coords, vel=vel_coords, time=time_coords, restfreq=restfreq)


def _add_common_attrs(
    xds: xr.Dataset,
    restfreq: float,
    spectral_reference: str,
    direction_reference: str,
    phase_center: Union[List[float], np.ndarray],
    cell_size: Union[List[float], np.ndarray],
    projection: str,
) -> xr.Dataset:
    xds.time.attrs = {"format": "MJD", "scale": "UTC", "units": "d"}
    freq_vals = np.array(xds.frequency)
    xds.frequency.attrs = {
        "rest_frequency": make_quantity(restfreq, "Hz"),
        "frame": spectral_reference.upper(),
        "units": "Hz",
        "wave_unit": "mm",
        # "crval": chan_coords[len(chan_coords) // 2],
        "crval": freq_vals[len(freq_vals) // 2].item(),
        "cdelt": (freq_vals[1] - freq_vals[0] if len(freq_vals) > 1 else 1000.0),
        "pc": 1.0,
    }
    xds.velocity.attrs = {"doppler_type": "RADIO", "units": "m/s"}
    xds.attrs = {
        "direction": {
            "reference": {
                "type": "sky_coord",
                "frame": direction_reference,
                "equinox": "J2000",
                "value": list(phase_center),
                "units": ["rad", "rad"],
            },
            "longpole": make_quantity(np.pi, "rad"),
            "latpole": make_quantity(0.0, "rad"),
            "pc": np.array([[1.0, 0.0], [0.0, 1.0]]),
            "projection": projection,
            "projection_parameters": [0.0, 0.0],
        },
        "active_mask": "",
        "beam": None,
        "object_name": "",
        "obsdate": {
            "type": "time",
            "scale": "UTC",
            "format": "MJD",
            "value": np.array(xds.time)[0],
            "units": "d",
        },
        "observer": "Karl Jansky",
        "pointing_center": {"value": list(phase_center), "initial": True},
        "description": "",
        "telescope": {
            "name": "ALMA",
            "position": {
                "type": "position",
                "ellipsoid": "GRS80",
                "units": ["rad", "rad", "m"],
                "value": [-1.1825465955049892, -0.3994149869262738, 6379946.01326443],
            },
        },
        "history": None,
    }
    return xds


def _make_empty_sky_image(
    phase_center: Union[list, np.ndarray],
    image_size: Union[list, np.ndarray],
    cell_size: Union[list, np.ndarray],
    chan_coords: Union[list, np.ndarray],
    pol_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
    direction_reference: str,
    projection: str,
    spectral_reference: str,
    do_sky_coords: bool,
) -> xr.Dataset:
    _input_checks(phase_center, image_size, cell_size)
    some_coords = _make_coords(chan_coords, time_coords)
    if do_sky_coords:
        long, lat = _compute_world_sph_dims(
            projection=projection,
            shape=image_size,
            ctype=["RA", "Dec"],
            crpix=[image_size[0] // 2, image_size[1] // 2],
            crval=phase_center,
            cdelt=[-abs(cell_size[0]), abs(cell_size[1])],
            cunit=["rad", "rad"],
        )["value"]
    # l follows RA as far as increasing/decreasing, see AIPS Meme 27, change in alpha
    # definition three lines below Figure 2 and the first of the pair of equations 10.
    l_coords = [
        (i - image_size[0] // 2) * (-1) * abs(cell_size[0])
        for i in range(image_size[0])
    ]
    m_coords = [
        (i - image_size[1] // 2) * abs(cell_size[1]) for i in range(image_size[1])
    ]
    coords = {
        "time": some_coords["time"],
        "polarization": pol_coords,
        "frequency": some_coords["chan"],
        "velocity": ("frequency", some_coords["vel"]),
        "l": l_coords,
        "m": m_coords,
    }
    if do_sky_coords:
        coords["right_ascension"] = (("l", "m"), long)
        coords["declination"] = (("l", "m"), lat)
    xds = xr.Dataset(coords=coords)
    attr_note = _l_m_attr_notes()
    xds.l.attrs = {
        "type": "quantity",
        "crval": 0.0,
        "cdelt": -abs(cell_size[0]),
        "units": "rad",
        "type": "quantity",
        "note": attr_note["l"],
    }
    xds.m.attrs = {
        "type": "quantity",
        "crval": 0.0,
        "cdelt": abs(cell_size[1]),
        "units": "rad",
        "type": "quantity",
        "note": attr_note["m"],
    }
    _add_common_attrs(
        xds,
        some_coords["restfreq"],
        spectral_reference,
        direction_reference,
        phase_center,
        cell_size,
        projection,
    )
    return xds


def _make_empty_aperture_image(
    phase_center: Union[list, np.ndarray],
    image_size: Union[list, np.ndarray],
    sky_image_cell_size: Union[list, np.ndarray],
    chan_coords: Union[list, np.ndarray],
    pol_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
    direction_reference: str,
    projection: str,
    spectral_reference: str,
) -> xr.Dataset:
    _input_checks(phase_center, image_size, sky_image_cell_size)
    some_coords = _make_coords(chan_coords, time_coords)
    im_size_wave = 1 / np.array(sky_image_cell_size)
    uv_cell_size = im_size_wave / np.array(image_size)
    u_coords = [
        (i - image_size[0] // 2) * abs(uv_cell_size[0]) for i in range(image_size[0])
    ]
    v_coords = [
        (i - image_size[1] // 2) * abs(uv_cell_size[1]) for i in range(image_size[1])
    ]
    coords = {
        "time": some_coords["time"],
        "polarization": pol_coords,
        "frequency": some_coords["chan"],
        "velocity": ("frequency", some_coords["vel"]),
        "u": u_coords,
        "v": v_coords,
    }
    xds = xr.Dataset(coords=coords)
    xds.u.attrs = {
        "crval": 0.0,
        "cdelt": abs(uv_cell_size[0]),
    }
    xds.v.attrs = {
        "crval": 0.0,
        "cdelt": abs(uv_cell_size[1]),
    }
    _add_common_attrs(
        xds,
        some_coords["restfreq"],
        spectral_reference,
        direction_reference,
        phase_center,
        sky_image_cell_size,
        projection,
    )
    return xds


def _make_empty_lmuv_image(
    phase_center: Union[list, np.ndarray],
    image_size: Union[list, np.ndarray],
    sky_image_cell_size: Union[list, np.ndarray],
    chan_coords: Union[list, np.ndarray],
    pol_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
    direction_reference: str,
    projection: str,
    spectral_reference: str,
    do_sky_coords: bool,
) -> xr.Dataset:
    _input_checks(phase_center, image_size, sky_image_cell_size)
    some_coords = _make_coords(chan_coords, time_coords)
    if do_sky_coords:
        long, lat = _compute_world_sph_dims(
            projection=projection,
            shape=image_size,
            ctype=["RA", "Dec"],
            crpix=[image_size[0] // 2, image_size[1] // 2],
            crval=phase_center,
            cdelt=[-abs(sky_image_cell_size[0]), abs(sky_image_cell_size[1])],
            cunit=["rad", "rad"],
        )["value"]
    # L follows RA as far as increasing/decreasing, see AIPS Meme 27, change in alpha
    # definition three lines below Figure 2 and the first of the pair of equations 10.
    l_coords = [
        (i - image_size[0] // 2) * (-1) * abs(sky_image_cell_size[0])
        for i in range(image_size[0])
    ]
    m_coords = [
        (i - image_size[1] // 2) * abs(sky_image_cell_size[1])
        for i in range(image_size[1])
    ]
    # im_size_wave = 1 / np.array(sky_image_cell_size)
    # uv_cell_size = im_size_wave / np.array(image_size)
    uv_cell_size = 1 / (np.array(image_size) * np.array(sky_image_cell_size))
    u_coords = [
        (i - image_size[0] // 2) * abs(uv_cell_size[0]) for i in range(image_size[0])
    ]
    v_coords = [
        (i - image_size[1] // 2) * abs(uv_cell_size[1]) for i in range(image_size[1])
    ]
    coords = {
        "time": some_coords["time"],
        "polarization": pol_coords,
        "frequency": some_coords["chan"],
        "velocity": ("frequency", some_coords["vel"]),
        "l": l_coords,
        "m": m_coords,
        "u": u_coords,
        "v": v_coords,
    }
    if do_sky_coords:
        coords["right_ascension"] = (("l", "m"), long)
        coords["declination"] = (("l", "m"), lat)
    xds = xr.Dataset(coords=coords)
    attr_note = _l_m_attr_notes()
    xds.l.attrs = {
        "type": "quantity",
        "crval": 0.0,
        "cdelt": -abs(sky_image_cell_size[0]),
        "units": "rad",
        "note": attr_note["l"],
    }
    xds.m.attrs = {
        "type": "quantity",
        "crval": 0.0,
        "cdelt": abs(sky_image_cell_size[1]),
        "units": "rad",
        "note": attr_note["m"],
    }
    xds.u.attrs = {
        "crval": 0.0,
        "cdelt": abs(uv_cell_size[0]),
    }
    xds.v.attrs = {
        "crval": 0.0,
        "cdelt": abs(uv_cell_size[1]),
    }
    _add_common_attrs(
        xds,
        some_coords["restfreq"],
        spectral_reference,
        direction_reference,
        phase_center,
        sky_image_cell_size,
        projection,
    )
    return xds
